import dataclasses
from itertools import product

import numpy as np
import numpy.testing
import pytest
import xarray as xr

from bioimageio.core import statistical_measures
from bioimageio.core.prediction_pipeline._measure_groups import get_measure_groups
from bioimageio.core.prediction_pipeline._utils import PER_DATASET, PER_SAMPLE
from bioimageio.core.statistical_measures import Mean, Percentile, Std, Var


@pytest.mark.parametrize("name_axes", product(["mean", "var", "std"], [None, ("x", "y")]))
def test_individual_normal_measure(name_axes):
    name, axes = name_axes
    measure = getattr(statistical_measures, name.title())(axes=axes)
    data = xr.DataArray(np.random.random((5, 6, 3)), dims=("x", "y", "c"))

    expected = getattr(data, name)(dim=axes)
    actual = measure.compute(data)
    xr.testing.assert_allclose(expected, actual)


@pytest.mark.parametrize("axes_n", product([None, ("x", "y")], [0, 10, 50, 100]))
def test_individual_percentile_measure(axes_n):
    axes, n = axes_n
    measure = statistical_measures.Percentile(axes=axes, n=n)
    data = xr.DataArray(np.random.random((5, 6, 3)), dims=("x", "y", "c"))

    expected = data.quantile(q=n / 100, dim=axes)
    actual = measure.compute(data)
    xr.testing.assert_allclose(expected, actual)


@pytest.mark.parametrize(
    "measures_mode",
    product(
        [
            {"t1": {Mean()}, "t2": {Mean(), Std()}},
            {"t1": {Mean(), Var(), Std()}, "t2": {Std(axes=("x", "y"))}},
            {"t1": {Mean(axes=("x", "y"))}, "t2": {Mean(), Std(axes=("x", "y"))}},
            {
                "t1": {Percentile(n=10), Percentile(n=35), Percentile(n=10, axes=("x", "y"))},
                "t2": {Percentile(n=10, axes=("x", "y")), Percentile(n=35, axes=("x", "y")), Percentile(n=10)},
            },
        ],
        [PER_SAMPLE, PER_DATASET],
    ),
)
def test_measure_groups(measures_mode):
    measures, mode = measures_mode

    def get_sample():
        return {
            "t1": xr.DataArray(np.random.random((2, 500, 600, 3)), dims=("b", "x", "y", "c")),
            "t2": xr.DataArray(np.random.random((1, 500, 600)), dims=("c", "x", "y")),
        }

    sample = get_sample()
    dataset_seq = [sample, get_sample()]
    dataset_full = {tn: xr.concat([s[tn] for s in dataset_seq], dim="dataset") for tn in sample.keys()}

    # compute independently
    expected = {}
    for tn, ms in measures.items():
        for m in ms:
            if mode == PER_SAMPLE:
                expected[(tn, m)] = m.compute(sample[tn])
            elif mode == PER_DATASET:
                if m.axes is None:
                    m_d = m
                else:
                    m_d = dataclasses.replace(m, axes=("dataset",) + m.axes)

                expected[(tn, m)] = m_d.compute(dataset_full[tn])
            else:
                raise NotImplementedError(mode)

    groups = get_measure_groups({mode: measures})[mode]
    actual = {}
    for g in groups:
        if mode == PER_SAMPLE:
            res = g.compute(sample)
        elif mode == PER_DATASET:
            for s in dataset_seq:
                g.update_with_sample(s)

            res = g.finalize()
        else:
            raise NotImplementedError(mode)

        for tn, vs in res.items():
            for m, v in vs.items():
                actual[(tn, m)] = v

    # discard additionally computed measures by groups
    actual = {k: v for k, v in actual.items() if k in expected}

    for k in expected.keys():
        assert k in actual
        numpy.testing.assert_array_almost_equal(expected[k].data, actual[k].data, decimal=2)



import logging
import warnings
from typing import List, Optional

import onnxruntime as rt
import xarray as xr

from ._model_adapter import ModelAdapter

logger = logging.getLogger(__name__)


class ONNXModelAdapter(ModelAdapter):
    def _load(self, *, devices: Optional[List[str]] = None):
        self._internal_output_axes = [tuple(out.axes) for out in self.bioimageio_model.outputs]

        self._session = rt.InferenceSession(str(self.bioimageio_model.weights["onnx"].source))
        onnx_inputs = self._session.get_inputs()
        self._input_names = [ipt.name for ipt in onnx_inputs]

        if devices is not None:
            warnings.warn(f"Device management is not implemented for onnx yet, ignoring the devices {devices}")

    def _forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        assert len(input_tensors) == len(self._input_names)
        input_arrays = [ipt.data for ipt in input_tensors]
        result = self._session.run(None, dict(zip(self._input_names, input_arrays)))
        if not isinstance(result, (list, tuple)):
            result = []

        return [xr.DataArray(r, dims=axes) for r, axes in zip(result, self._internal_output_axes)]

    def _unload(self) -> None:
        warnings.warn("Device management is not implemented for onnx yet, cannot unload model")



import warnings
from typing import List, Optional, Sequence
from marshmallow import missing

# by default, we use the keras integrated with tensorflow
try:
    from tensorflow import keras
    import tensorflow as tf

    TF_VERSION = tf.__version__
except Exception:
    import keras

    TF_VERSION = None
import xarray as xr

from ._model_adapter import ModelAdapter


class KerasModelAdapter(ModelAdapter):
    def _load(self, *, devices: Optional[Sequence[str]] = None) -> None:
        model_tf_version = self.bioimageio_model.weights["keras_hdf5"].tensorflow_version
        if model_tf_version is missing:
            model_tf_version = None
        else:
            model_tf_version = (int(model_tf_version.major), int(model_tf_version.minor))

        if TF_VERSION is None or model_tf_version is None:
            warnings.warn("Could not check tensorflow versions. The prediction results may be wrong.")
        elif tuple(model_tf_version[:2]) != tuple(map(int, TF_VERSION.split(".")))[:2]:
            warnings.warn(
                f"Model tensorflow version {model_tf_version} does not match {TF_VERSION}."
                "The prediction results may be wrong"
            )

        # TODO keras device management
        if devices is not None:
            warnings.warn(f"Device management is not implemented for keras yet, ignoring the devices {devices}")

        weight_file = self.bioimageio_model.weights["keras_hdf5"].source
        self._model = keras.models.load_model(weight_file)
        self._output_axes = [tuple(out.axes) for out in self.bioimageio_model.outputs]

    def _unload(self) -> None:
        warnings.warn("Device management is not implemented for keras yet, cannot unload model")

    def _forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        result = self._model.predict(*input_tensors)
        if not isinstance(result, (tuple, list)):
            result = [result]

        assert len(result) == len(self._output_axes)
        return [xr.DataArray(r, dims=axes) for r, axes, in zip(result, self._output_axes)]



from typing import Dict, Set

import xarray as xr

from bioimageio.core.statistical_measures import Measure, MeasureValue

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

TensorName = str
FixedMode = Literal["fixed"]
SampleMode = Literal["per_sample"]
DatasetMode = Literal["per_dataset"]
Mode = Literal[FixedMode, SampleMode, DatasetMode]

FIXED: FixedMode = "fixed"
PER_SAMPLE: SampleMode = "per_sample"
PER_DATASET: DatasetMode = "per_dataset"
MODES: Set[Mode] = {FIXED, PER_SAMPLE, PER_DATASET}

Sample = Dict[TensorName, xr.DataArray]
RequiredMeasures = Dict[Literal[SampleMode, DatasetMode], Dict[TensorName, Set[Measure]]]
ComputedMeasures = Dict[Literal[SampleMode, DatasetMode], Dict[TensorName, Dict[Measure, MeasureValue]]]



import numpy as np
import xarray as xr


def test_binarize_postprocessing():
    from bioimageio.core.prediction_pipeline._processing import Binarize

    shape = (3, 32, 32)
    axes = ("c", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)

    threshold = 0.5
    exp = xr.DataArray(np_data > threshold, dims=axes)

    binarize = Binarize("data_name", threshold=threshold)
    res = binarize(data)
    xr.testing.assert_allclose(res, exp)



import numpy as np
import xarray as xr
from numpy.testing import assert_array_almost_equal

from bioimageio.core import load_resource_description
from bioimageio.core.resource_io.nodes import Model


def _test_prediction_pipeline(model_package, weight_format):
    from bioimageio.core.prediction_pipeline import create_prediction_pipeline

    bio_model = load_resource_description(model_package)
    assert isinstance(bio_model, Model)
    pp = create_prediction_pipeline(bioimageio_model=bio_model, weight_format=weight_format)

    inputs = [
        xr.DataArray(np.load(str(test_tensor)), dims=tuple(spec.axes))
        for test_tensor, spec in zip(bio_model.test_inputs, bio_model.inputs)
    ]
    outputs = pp.forward(*inputs)
    assert isinstance(outputs, list)

    expected_outputs = [
        xr.DataArray(np.load(str(test_tensor)), dims=tuple(spec.axes))
        for test_tensor, spec in zip(bio_model.test_outputs, bio_model.outputs)
    ]
    assert len(outputs) == len(expected_outputs)

    for out, exp in zip(outputs, expected_outputs):
        assert_array_almost_equal(out, exp, decimal=4)


def test_prediction_pipeline_torch(any_torch_model):
    _test_prediction_pipeline(any_torch_model, "pytorch_state_dict")


def test_prediction_pipeline_torchscript(any_torchscript_model):
    _test_prediction_pipeline(any_torchscript_model, "torchscript")


def test_prediction_pipeline_onnx(any_onnx_model):
    _test_prediction_pipeline(any_onnx_model, "onnx")


def test_prediction_pipeline_tensorflow(any_tensorflow_model):
    _test_prediction_pipeline(any_tensorflow_model, "tensorflow_saved_model_bundle")


def test_prediction_pipeline_keras(any_keras_model):
    _test_prediction_pipeline(any_keras_model, "keras_hdf5")



from typing import Dict, Iterable, Optional

from tqdm import tqdm

from bioimageio.core.statistical_measures import Measure
from ._measure_groups import MeasureGroups, MeasureValue, get_measure_groups
from ._utils import ComputedMeasures, PER_DATASET, PER_SAMPLE, RequiredMeasures, Sample, TensorName

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


class StatsState:
    """class to compute, hold and update dataset and sample statistics"""

    sample_count: int
    last_sample: Optional[Sample]
    measure_groups: MeasureGroups
    _n_start: int
    _n_stop: int
    _final_dataset_stats: Optional[Dict[TensorName, Dict[Measure, MeasureValue]]]

    def __init__(
        self,
        required_measures: RequiredMeasures,
        *,
        dataset: Iterable[Sample] = tuple(),
        update_dataset_stats_after_n_samples: Optional[int] = None,
        update_dataset_stats_for_n_samples: int = float("inf"),
    ):
        """iterates over dataset to compute dataset statistics (if required). The resulting dataset statistics are further updated with each new sample. A sample in this context may be a mini-batch.

        Args:
            required_measures: measures to be computed
            dataset: (partial) dataset to initialize dataset statistics with
            update_dataset_stats_after_n_samples: Update dataset statistics for new samples S_i if i > n.
                                                  (default: len(dataset))
                                                  This parameter allows to avoid weighting the first n processed
                                                  samples to count twice if they make up the given 'dataset'.
            update_dataset_stats_for_n_samples: stop updating dataset statistics with new samples S_i if
                                                i > for_n_samples (+ update_dataset_stats_after_n_samples)
        """
        self.required_measures = required_measures
        self.update_dataset_stats_after_n_samples = update_dataset_stats_after_n_samples
        self.update_dataset_stats_for_n_samples = update_dataset_stats_for_n_samples
        self.reset(dataset)

    def reset(self, dataset: Iterable[Sample]):
        self.sample_count = 0
        self.last_sample = None
        self._final_dataset_stats = None
        self.measure_groups = get_measure_groups(self.required_measures)

        len_dataset = 0
        if self.measure_groups[PER_DATASET]:
            for sample in tqdm(dataset, "computing dataset statistics"):
                len_dataset += 1
                self._update_dataset_measure_groups(sample)

        if self.update_dataset_stats_after_n_samples is None:
            self._n_start = len_dataset
        else:
            self._n_start = self.update_dataset_stats_after_n_samples

        self._n_stop = self._n_start + self.update_dataset_stats_for_n_samples

    def update_with_sample(self, sample: Sample):
        self.last_sample = sample
        self.sample_count += 1
        if self._n_start < self.sample_count <= self._n_stop:
            self._update_dataset_measure_groups(sample)

    def _update_dataset_measure_groups(self, sample: Sample):
        for mg in self.measure_groups[PER_DATASET]:
            mg.update_with_sample(sample)

    def compute_measures(self) -> ComputedMeasures:
        ret = {PER_SAMPLE: {}, PER_DATASET: {}}
        if self.last_sample is not None:
            for mg in self.measure_groups[PER_SAMPLE]:
                ret[PER_SAMPLE].update(mg.compute(self.last_sample))

        if self._final_dataset_stats is None:
            dataset_stats = {}
            for mg in self.measure_groups[PER_DATASET]:
                dataset_stats.update(mg.finalize())

            if self.sample_count > self._n_stop:
                # stop recomputing final dataset statistics
                self._final_dataset_stats = dataset_stats
        else:
            dataset_stats = self._final_dataset_stats

        ret[PER_DATASET] = dataset_stats
        return ret



import datetime
import hashlib
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from warnings import warn

import imageio
import numpy as np
import requests
import tifffile

import bioimageio.spec as spec
import bioimageio.spec.model as model_spec
from bioimageio.core import export_resource_package, load_raw_resource_description
from bioimageio.core.resource_io.nodes import URI
from bioimageio.spec.shared.raw_nodes import ImportableModule, ImportableSourceFile
from bioimageio.spec.shared import resolve_local_source, resolve_source

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore


#
# utility functions to build the spec from python
#


def _get_hash(path):
    with open(path, "rb") as f:
        data = f.read()
        return hashlib.sha256(data).hexdigest()


def _infer_weight_type(path):
    ext = os.path.splitext(path)[-1]
    if ext in (".pt", ".pth", ".torch"):
        return "pytorch_state_dict"
    elif ext == ".onnx":
        return "onnx"
    elif ext in (".hdf", ".hdf5", ".h5"):
        return "keras_hdf5"
    elif ext == ".zip":
        return "tensorflow_saved_model_bundle"
    elif ext == ".json":
        return "tensorflow_js"
    else:
        raise ValueError(f"Could not infer weight type from extension {ext} for weight file {path}")


def _get_pytorch_state_dict_weight_kwargs(architecture, model_kwargs, root):
    assert architecture is not None
    tmp_archtecture = None
    weight_kwargs = {"kwargs": model_kwargs} if model_kwargs else {}
    if ":" in architecture:
        # note: path itself might include : for absolute paths in windows
        *arch_file_parts, callable_name = architecture.replace("::", ":").split(":")
        arch_file = _ensure_local(":".join(arch_file_parts), root)
        arch = ImportableSourceFile(callable_name, arch_file)
        arch_hash = _get_hash(root / arch.source_file)
        weight_kwargs["architecture_sha256"] = arch_hash
    else:
        arch = spec.shared.fields.ImportableSource().deserialize(architecture)
        assert isinstance(arch, ImportableModule)

    weight_kwargs["architecture"] = arch
    return weight_kwargs, tmp_archtecture


def _get_attachments(attachments, root):
    assert isinstance(attachments, dict)
    if "files" in attachments:
        afiles = attachments["files"]
        if isinstance(afiles, str):
            afiles = [afiles]

        if isinstance(afiles, list):
            afiles = _ensure_local_or_url(afiles, root)
        else:
            raise TypeError(attachments)

        attachments["files"] = afiles
    return attachments


def _get_weights(
    original_weight_source,
    weight_type,
    root,
    architecture=None,
    model_kwargs=None,
    tensorflow_version=None,
    opset_version=None,
    pytorch_version=None,
    dependencies=None,
    attachments=None,
):
    weight_path = resolve_source(original_weight_source, root)
    if weight_type is None:
        weight_type = _infer_weight_type(weight_path)
    weight_hash = _get_hash(weight_path)

    weight_types = model_spec.raw_nodes.WeightsFormat
    weight_source = _ensure_local_or_url(original_weight_source, root)

    weight_kwargs = {"source": weight_source, "sha256": weight_hash}
    if attachments is not None:
        weight_kwargs["attachments"] = _get_attachments(attachments, root)
    if dependencies is not None:
        weight_kwargs["dependencies"] = _get_dependencies(dependencies, root)

    tmp_archtecture = None
    if weight_type == "pytorch_state_dict":
        # pytorch-state-dict -> we need an architecture definition
        pytorch_weight_kwargs, tmp_file = _get_pytorch_state_dict_weight_kwargs(architecture, model_kwargs, root)
        weight_kwargs.update(**pytorch_weight_kwargs)
        if pytorch_version is not None:
            weight_kwargs["pytorch_version"] = pytorch_version
        elif dependencies is None:
            warn(
                "You are building a pytorch model but have neither passed dependencies nor the pytorch_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.PytorchStateDictWeightsEntry(**weight_kwargs)

    elif weight_type == "onnx":
        if opset_version is not None:
            weight_kwargs["opset_version"] = opset_version
        elif dependencies is None:
            warn(
                "You are building an onnx model but have neither passed dependencies nor the opset_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.OnnxWeightsEntry(**weight_kwargs)

    elif weight_type == "torchscript":
        if pytorch_version is not None:
            weight_kwargs["pytorch_version"] = pytorch_version
        elif dependencies is None:
            warn(
                "You are building a pytorch model but have neither passed dependencies nor the pytorch_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.TorchscriptWeightsEntry(**weight_kwargs)

    elif weight_type == "keras_hdf5":
        if tensorflow_version is not None:
            weight_kwargs["tensorflow_version"] = tensorflow_version
        elif dependencies is None:
            warn(
                "You are building a keras model but have neither passed dependencies nor the tensorflow_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.KerasHdf5WeightsEntry(**weight_kwargs)

    elif weight_type == "tensorflow_saved_model_bundle":
        if tensorflow_version is not None:
            weight_kwargs["tensorflow_version"] = tensorflow_version
        elif dependencies is None:
            warn(
                "You are building a tensorflow model but have neither passed dependencies nor the tensorflow_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.TensorflowSavedModelBundleWeightsEntry(**weight_kwargs)

    elif weight_type == "tensorflow_js":
        if tensorflow_version is not None:
            weight_kwargs["tensorflow_version"] = tensorflow_version
        elif dependencies is None:
            warn(
                "You are building a tensorflow model but have neither passed dependencies nor the tensorflow_version."
                "It may not be possible to create an environmnet where your model can be used."
            )
        weights = model_spec.raw_nodes.TensorflowJsWeightsEntry(**weight_kwargs)

    elif weight_type in weight_types:
        raise ValueError(f"Weight type {weight_type} is not supported yet in 'build_spec'")
    else:
        raise ValueError(f"Invalid weight type {weight_type}, expect one of {weight_types}")

    return {weight_type: weights}, tmp_archtecture


def _get_data_range(data_range, dtype):
    if data_range is None:
        if np.issubdtype(np.dtype(dtype), np.integer):
            min_, max_ = np.iinfo(dtype).min, np.iinfo(dtype).max
        # for floating point numbers we assume valid range from -inf to inf
        elif np.issubdtype(np.dtype(dtype), np.floating):
            min_, max_ = -np.inf, np.inf
        elif np.issubdtype(np.dtype(dtype), np.bool):
            min_, max_ = 0, 1
        else:
            raise RuntimeError(f"Cannot derived data range for dtype {dtype}")
        data_range = (min_, max_)
    assert isinstance(data_range, (tuple, list)), type(data_range)
    assert len(data_range) == 2
    return data_range


def _get_input_tensor(path, name, step, min_shape, data_range, axes, preprocessing):
    test_in = np.load(path)
    shape = test_in.shape
    if step is None:
        assert min_shape is None
        shape_description = shape
    else:
        shape_description = {"min": shape if min_shape is None else min_shape, "step": step}

    data_range = _get_data_range(data_range, test_in.dtype)
    kwargs = {}
    if preprocessing is not None:
        kwargs["preprocessing"] = preprocessing

    inputs = model_spec.raw_nodes.InputTensor(
        name="input" if name is None else name,
        data_type=str(test_in.dtype),
        axes=axes,
        shape=shape_description,
        data_range=data_range,
        **kwargs,
    )
    return inputs


def _get_output_tensor(path, name, reference_tensor, scale, offset, axes, data_range, postprocessing, halo):
    test_out = np.load(path)
    shape = test_out.shape
    if reference_tensor is None:
        assert scale is None
        assert offset is None
        shape_description = shape
    else:
        assert scale is not None
        assert offset is not None
        shape_description = {"reference_tensor": reference_tensor, "scale": scale, "offset": offset}

    data_range = _get_data_range(data_range, test_out.dtype)
    kwargs = {}
    if postprocessing is not None:
        kwargs["postprocessing"] = postprocessing
    if halo is not None:
        kwargs["halo"] = halo

    outputs = model_spec.raw_nodes.OutputTensor(
        name="output" if name is None else name,
        data_type=str(test_out.dtype),
        axes=axes,
        data_range=data_range,
        shape=shape_description,
        **kwargs,
    )
    return outputs


def _build_cite(cite: List[Dict[str, str]]):
    citation_list = []
    for entry in cite:
        if "doi" in entry:
            spec_entry = spec.rdf.raw_nodes.CiteEntry(text=entry["text"], doi=entry["doi"])
        elif "url" in entry:
            spec_entry = spec.rdf.raw_nodes.CiteEntry(text=entry["text"], url=entry["url"])
        else:
            raise ValueError(f"Expect one of doi or url in citation enrty {entry}")
        citation_list.append(spec_entry)
    return citation_list


def _get_dependencies(dependencies, root):
    if isinstance(dependencies, Path) or ":" not in dependencies:
        manager = "conda"
        path = dependencies
    else:
        manager, path = dependencies.split(":")

    return model_spec.raw_nodes.Dependencies(manager=manager, file=_ensure_local(path, root))


def _get_deepimagej_macro(name, kwargs, export_folder):

    # macros available in deepimagej
    macro_names = ("binarize", "scale_linear", "scale_range", "zero_mean_unit_variance")
    if name == "scale_linear":
        macro = "scale_linear.ijm"
        replace = {"gain": kwargs["gain"], "offset": kwargs["offset"]}

    elif name == "scale_range":
        macro = "per_sample_scale_range.ijm"
        replace = {"min_precentile": kwargs["min_percentile"], "max_percentile": kwargs["max_percentile"]}

    elif name == "zero_mean_unit_variance":
        mode = kwargs["mode"]
        if mode == "fixed":
            macro = "fixed_zero_mean_unit_variance.ijm"
            replace = {"paramMean": kwargs["mean"], "paramStd": kwargs["std"]}
        else:
            macro = "zero_mean_unit_variance.ijm"
            replace = {}

    elif name == "binarize":
        macro = "binarize.ijm"
        replace = {"optimalThreshold": kwargs["threshold"]}

    else:
        raise ValueError(f"Macro {name} is not available, must be one of {macro_names}.")

    url = f"https://raw.githubusercontent.com/deepimagej/imagej-macros/master/bioimage.io/{macro}"

    path = os.path.join(export_folder, macro)
    # use https://github.com/bioimage-io/core-bioimage-io-python/blob/main/bioimageio/core/resource_io/utils.py#L267
    # instead if the implementation is update s.t. an output path is accepted
    with requests.get(url, stream=True) as r:
        text = r.text
        if text.startswith("4"):
            raise RuntimeError(f"An error occured when downloading {url}: {r.text}")
        with open(path, "w") as f:
            f.write(r.text)

    # replace the kwargs in the macro file
    if replace:
        lines = []
        with open(path) as f:
            for line in f:
                kwarg = [kwarg for kwarg in replace if line.startswith(kwarg)]
                if kwarg:
                    assert len(kwarg) == 1
                    kwarg = kwarg[0]
                    # each kwarg should only be replaced ones
                    val = replace.pop(kwarg)
                    lines.append(f"{kwarg} = {val};\n")
                else:
                    lines.append(line)

        with open(path, "w") as f:
            for line in lines:
                f.write(line)

    return {"spec": "ij.IJ::runMacroFile", "kwargs": macro}


def _get_deepimagej_config(
    export_folder, test_inputs, test_outputs, input_axes, output_axes, pixel_sizes, preprocessing, postprocessing
):
    assert len(test_inputs) == len(test_outputs) == 1, "deepimagej config only valid for single input/output"

    if any(preproc is not None for preproc in preprocessing):
        assert len(preprocessing) == 1
        preprocess_ij = [
            _get_deepimagej_macro(preproc["name"], preproc["kwargs"], export_folder) for preproc in preprocessing[0]
        ]
        attachments = [preproc["kwargs"] for preproc in preprocess_ij]
    else:
        preprocess_ij = [{"spec": None}]
        attachments = None

    if any(postproc is not None for postproc in postprocessing):
        assert len(postprocessing) == 1
        postprocess_ij = [
            _get_deepimagej_macro(postproc["name"], postproc["kwargs"], export_folder) for postproc in postprocessing[0]
        ]
        if attachments is None:
            attachments = [postproc["kwargs"] for postproc in postprocess_ij]
        else:
            attachments.extend([postproc["kwargs"] for postproc in postprocess_ij])
    else:
        postprocess_ij = [{"spec": None}]

    def get_size(fname, axes):
        shape = np.load(export_folder / fname).shape
        assert len(shape) == len(axes)
        shape = [sh for sh, ax in zip(shape, axes) if ax != "b"]
        axes = [ax for ax in axes if ax != "b"]
        # the shape for deepij is always given as xyzc
        if len(shape) == 3:
            axes_ij = "xyc"
        else:
            axes_ij = "xyzc"
        assert set(axes) == set(axes_ij)
        axis_permutation = [axes_ij.index(ax) for ax in axes]
        shape = [shape[permut] for permut in axis_permutation]
        if len(shape) == 3:
            shape = shape[:2] + [1] + shape[-1:]
        assert len(shape) == 4
        return " x ".join(map(str, shape))

    # deepimagej always expexts a pixel size for the z axis
    pixel_sizes_ = [pix_size if "z" in pix_size else dict(z=1.0, **pix_size) for pix_size in pixel_sizes]

    test_info = {
        "inputs": [
            {"name": in_path, "size": get_size(in_path, axes), "pixel_size": pix_size}
            for in_path, axes, pix_size in zip(test_inputs, input_axes, pixel_sizes_)
        ],
        "outputs": [
            {"name": out_path, "type": "image", "size": get_size(out_path, axes)}
            for out_path, axes in zip(test_outputs, output_axes)
        ],
        "memory_peak": None,
        "runtime": None,
    }

    config = {
        "prediction": {"preprocess": preprocess_ij, "postprocess": postprocess_ij},
        "test_information": test_info,
        # other stuff deepimagej needs
        "pyramidal_model": False,
        "allow_tiling": True,
        "model_keys": None,
    }
    return {"deepimagej": config}, [Path(a) for a in attachments]


def _write_sample_data(input_paths, output_paths, input_axes, output_axes, pixel_sizes, export_folder: Path):
    def write_im(path, im, axes, pixel_size=None):
        assert len(axes) == im.ndim, f"{len(axes), {im.ndim}}"
        assert im.ndim in (4, 5), f"{im.ndim}"

        # convert the image to expects (Z)CYX axis order
        if im.ndim == 4:
            assert set(axes) == {"b", "x", "y", "c"}, f"{axes}"
            resolution_axes_ij = "cyxb"
        else:
            assert set(axes) == {"b", "x", "y", "z", "c"}, f"{axes}"
            resolution_axes_ij = "bzcyx"

        def addMissingAxes(im_axes):
            needed_axes = ["b", "c", "x", "y", "z", "s"]
            for ax in needed_axes:
                if ax not in im_axes:
                    im_axes += ax
            return im_axes

        axes_ij = "bzcyxs"
        # Expand the image to ImageJ dimensions
        im = np.expand_dims(im, axis=tuple(range(len(axes), len(axes_ij))))

        axis_permutation = tuple(addMissingAxes(axes).index(ax) for ax in axes_ij)
        im = im.transpose(axis_permutation)

        if pixel_size is None:
            resolution = None
        else:
            spatial_axes = list(set(resolution_axes_ij) - set("bc"))
            resolution = tuple(1.0 / pixel_size[ax] for ax in resolution_axes_ij if ax in spatial_axes)
        # does not work for double
        if np.dtype(im.dtype) == np.dtype("float64"):
            im = im.astype("float32")
        tifffile.imwrite(path, im, imagej=True, resolution=resolution)

    sample_in_paths = []
    for i, (in_path, axes) in enumerate(zip(input_paths, input_axes)):
        inp = np.load(export_folder / in_path)
        sample_in_path = export_folder / f"sample_input_{i}.tif"
        pixel_size = None if pixel_sizes is None else pixel_sizes[i]
        write_im(sample_in_path, inp, axes, pixel_size)
        sample_in_paths.append(sample_in_path)

    sample_out_paths = []
    for i, (out_path, axes) in enumerate(zip(output_paths, output_axes)):
        outp = np.load(export_folder / out_path)
        sample_out_path = export_folder / f"sample_output_{i}.tif"
        write_im(sample_out_path, outp, axes)
        sample_out_paths.append(sample_out_path)

    return [Path(p.name) for p in sample_in_paths], [Path(p.name) for p in sample_out_paths]


# create better cover images for 3d data and non-image outputs
def _generate_covers(in_path, out_path, input_axes, output_axes, root):
    def normalize(data, axis, eps=1e-7):
        data = data.astype("float32")
        data -= data.min(axis=axis, keepdims=True)
        data /= data.max(axis=axis, keepdims=True) + eps
        return data

    def to_image(data, data_axes):
        assert data.ndim in (4, 5)

        # transpose the data to "bczyx" / "bcyx" order
        axes = "bczyx" if data.ndim == 5 else "bcyx"
        assert set(data_axes) == set(axes)
        if axes != data_axes:
            ax_permutation = tuple(data_axes.index(ax) for ax in axes)
            data = data.transpose(ax_permutation)

        # select single image with channels from the data
        if data.ndim == 5:
            z0 = data.shape[2] // 2
            data = data[0, :, z0]
        else:
            data = data[0, :]

        # normalize the data and map to 8 bit
        data = normalize(data, axis=(1, 2))
        data = (data * 255).astype("uint8")
        return data

    cover_path = os.path.join(root, "cover.png")
    input_, output = np.load(in_path), np.load(out_path)

    input_ = to_image(input_, input_axes)
    # this is not image data so we only save the input image
    if output.ndim < 4:
        imageio.imwrite(cover_path, input_.transpose((1, 2, 0)))
        return [_ensure_local(cover_path, root)]
    output = to_image(output, output_axes)

    chan_in = input_.shape[0]
    # make sure the input is rgb
    if chan_in == 1:  # single channel -> repeat it 3 times
        input_ = np.repeat(input_, 3, axis=0)
    elif chan_in != 3:  # != 3 channels -> take first channe and repeat it 3 times
        input_ = np.repeat(input_[0:1], 3, axis=0)

    im_shape = input_.shape[1:]
    # we just save the input image if the shapes don't agree
    if im_shape != output.shape[1:]:
        imageio.imwrite(cover_path, input_.transpose((1, 2, 0)))
        return [_ensure_local(cover_path, root)]

    def diagonal_split(im0, im1):
        assert im0.shape[0] == im1.shape[0] == 3
        n, m = im_shape
        out = np.ones((3, n, m), dtype="uint8")
        for c in range(3):
            outc = np.tril(im0[c])
            mask = outc == 0
            outc[mask] = np.triu(im1[c])[mask]
            out[c] = outc
        return out

    def grid_im(im0, im1):
        ims_per_row = 3
        n_chan = im1.shape[0]
        n_images = n_chan + 1
        n_rows = int(np.ceil(float(n_images) / ims_per_row))

        n, m = im_shape
        x, y = ims_per_row * n, n_rows * m
        out = np.zeros((3, y, x))
        images = [im0] + [np.repeat(im1[i : i + 1], 3, axis=0) for i in range(n_chan)]

        i, j = 0, 0
        for im in images:
            x0, x1 = i * n, (i + 1) * n
            y0, y1 = j * m, (j + 1) * m
            out[:, y0:y1, x0:x1] = im

            i += 1
            if i == ims_per_row:
                i = 0
                j += 1

        return out

    chan_out = output.shape[0]
    if chan_out == 1:  # single prediction channel: create diagonal split
        im = diagonal_split(input_, np.repeat(output, 3, axis=0))
    elif chan_out == 3:  # three prediction channel: create diagonal split with rgb
        im = diagonal_split(input_, output)
    else:  # otherwise create grid image
        im = grid_im(input_, output)

    # to channel last
    imageio.imwrite(cover_path, im.transpose((1, 2, 0)))
    return [_ensure_local(cover_path, root)]


def _ensure_local(source: Union[Path, URI, str, list], root: Path) -> Union[Path, URI, list]:
    """ensure source is local relative path in root"""
    if isinstance(source, list):
        return [_ensure_local(s, root) for s in source]

    local_source = resolve_source(source, root)
    local_source = resolve_source(local_source, root, root / local_source.name)
    return local_source.relative_to(root)


def _ensure_local_or_url(source: Union[Path, URI, str, list], root: Path) -> Union[Path, URI, list]:
    """ensure source is remote URI or local relative path in root"""
    if isinstance(source, list):
        return [_ensure_local_or_url(s, root) for s in source]

    local_source = resolve_local_source(source, root)
    if not isinstance(local_source, URI):
        local_source = resolve_local_source(local_source, root, root / local_source.name)
    return local_source.relative_to(root)


def build_model(
    # model or tensor specific and required
    weight_uri: str,
    test_inputs: List[Union[str, Path]],
    test_outputs: List[Union[str, Path]],
    input_axes: List[str],
    output_axes: List[str],
    # general required
    name: str,
    description: str,
    authors: List[Dict[str, str]],
    tags: List[Union[str, Path]],
    documentation: Union[str, Path],
    cite: List[Dict[str, str]],
    output_path: Union[str, Path],
    # model specific optional
    architecture: Optional[str] = None,
    model_kwargs: Optional[Dict[str, Union[int, float, str]]] = None,
    weight_type: Optional[str] = None,
    sample_inputs: Optional[List[str]] = None,
    sample_outputs: Optional[List[str]] = None,
    # tensor specific
    input_names: Optional[List[str]] = None,
    input_step: Optional[List[List[int]]] = None,
    input_min_shape: Optional[List[List[int]]] = None,
    input_data_range: Optional[List[List[Union[int, str]]]] = None,
    output_names: Optional[List[str]] = None,
    output_reference: Optional[List[str]] = None,
    output_scale: Optional[List[List[int]]] = None,
    output_offset: Optional[List[List[int]]] = None,
    output_data_range: Optional[List[List[Union[int, str]]]] = None,
    halo: Optional[List[List[int]]] = None,
    preprocessing: Optional[List[List[Dict[str, Dict[str, Union[int, float, str]]]]]] = None,
    postprocessing: Optional[List[List[Dict[str, Dict[str, Union[int, float, str]]]]]] = None,
    pixel_sizes: Optional[List[Dict[str, float]]] = None,
    # general optional
    maintainers: Optional[List[Dict[str, str]]] = None,
    license: Optional[str] = None,
    covers: Optional[List[str]] = None,
    git_repo: Optional[str] = None,
    attachments: Optional[Dict[str, Union[str, List[str]]]] = None,
    packaged_by: Optional[List[str]] = None,
    run_mode: Optional[str] = None,
    parent: Optional[Dict[str, str]] = None,
    config: Optional[Dict[str, Any]] = None,
    dependencies: Optional[Union[Path, str]] = None,
    links: Optional[List[str]] = None,
    training_data: Optional[Dict[str, str]] = None,
    root: Optional[Union[Path, str]] = None,
    add_deepimagej_config: bool = False,
    tensorflow_version: Optional[str] = None,
    opset_version: Optional[int] = None,
    pytorch_version: Optional[str] = None,
    weight_attachments: Optional[Dict[str, Union[str, List[str]]]] = None,
):
    """Create a zipped bioimage.io model.

    Example usage:
    ```
    from pathlib import Path
    import bioimageio.spec as spec
    import bioimageio.core.build_spec as build_spec
    model_spec = build_spec.build_model(
        weight_uri="test_weights.pt",
        test_inputs=["./test_inputs"],
        test_outputs=["./test_outputs"],
        input_axes=["bcyx"],
        output_axes=["bcyx"],
        name="my-model",
        description="My very fancy model.",
        authors=[{"name": "John Doe", "affiliation": "My Institute"}],
        tags=["segmentation", "light sheet data"],
        license="CC-BY-4.0",
        documentation="./documentation.md",
        cite=[{"text": "Ronneberger et al. U-Net", "doi": "10.1007/978-3-319-24574-4_28"}],
        output_path="my-model.zip"
    )
    ```

    Args:
        weight_uri: the url or relative local file path to the weight file for this model.
        test_inputs: list of test input files stored in numpy format.
        test_outputs: list of test outputs corresponding to test_inputs, stored in numpy format.
        input_axes: axis names of the input tensors.
        output_axes: axiss names of the output tensors.
        name: name of this model.
        description: short description of this model.
        authors: the authors of this model.
        tags: list of tags for this model.
        documentation: relative file path to markdown documentation for this model.
        cite: references for this model.
        output_path: where to save the zipped model package.
        architecture: the file with the source code for the model architecture and the corresponding class.
            Only required for models with pytorch_state_dict weight format.
        model_kwargs: the keyword arguments for the model class.
            Only required for models with pytorch_state_dict weight format.
        weight_type: the type of the weights.
        sample_inputs: list of sample inputs to demonstrate the model performance.
        sample_outputs: list of sample outputs corresponding to sample_inputs.
        input_names: names of the input tensors.
        input_step: minimal valid increase of the input tensor shape.
        input_min_shape: minimal input tensor shape.
        input_data_range: valid data range for the input tensor.
        output_names: names of the output tensors.
        output_reference: name of the input reference tensor used to cimpute the output tensor shape.
        output_scale: multiplicative factor to compute the output tensor shape.
        output_offset: additive term to compute the output tensor shape.
        output_data_range: valid data range for the output tensor.
        halo: halo to be cropped from the output tensor.
        preprocessing: list of preprocessing operations for the input.
        postprocessing: list of postprocessing operations for the output.
        pixel_sizes: the pixel sizes for the input tensors, only for spatial axes.
            This information is currently only used by deepimagej, but will be added to the spec soon.
        license: the license for this model. By default CC-BY-4.0 will be set as license.
        covers: list of file paths for cover images.
            By default a cover will be generated from the input and output data.
        git_repo: reference git repository for this model.
        attachments: list of additional files to package with the model.
        packaged_by: list of authors that have packaged this model.
        run_mode: custom run mode for this model.
        parent: id of the parent model from which this model is derived and sha256 of the corresponding rdf file.
        config: custom configuration for this model.
        dependencies: relative path to file with dependencies for this model.
        training_data: the training data for this model, either id for a bioimageio dataset or a dataset spec.
        root: optional root path for relative paths. This can be helpful when building a spec from another model spec.
        add_deepimagej_config: add the deepimagej config to the model.
        tensorflow_version: the tensorflow version for this model. Only for tensorflow or keras weights.
        opset_version: the opset version for this model. Only for onnx weights.
        pytorch_version: the pytorch version for this model. Only for pytoch_state_dict or torchscript weights.
        weight_attachments: extra weight specific attachments.
    """
    assert architecture is None or isinstance(architecture, str)
    if root is None:
        root = "."
    root = Path(root)

    if attachments is not None:
        attachments = _get_attachments(attachments, root)

    #
    # generate the model specific fields
    #

    assert len(test_inputs)
    assert len(test_outputs)
    test_inputs = _ensure_local_or_url(test_inputs, root)
    test_outputs = _ensure_local_or_url(test_outputs, root)

    n_inputs = len(test_inputs)
    if input_names is None:
        input_names = [f"input{i}" for i in range(n_inputs)]
    else:
        assert len(input_names) == len(test_inputs)

    input_step = n_inputs * [None] if input_step is None else input_step
    input_min_shape = n_inputs * [None] if input_min_shape is None else input_min_shape
    input_data_range = n_inputs * [None] if input_data_range is None else input_data_range
    preprocessing = n_inputs * [None] if preprocessing is None else preprocessing

    inputs = [
        _get_input_tensor(root / test_in, name, step, min_shape, data_range, axes, preproc)
        for test_in, name, step, min_shape, axes, data_range, preproc in zip(
            test_inputs, input_names, input_step, input_min_shape, input_axes, input_data_range, preprocessing
        )
    ]

    n_outputs = len(test_outputs)
    if output_names is None:
        output_names = [f"output{i}" for i in range(n_outputs)]
    else:
        assert len(output_names) == len(test_outputs)

    output_reference = n_outputs * [None] if output_reference is None else output_reference
    output_scale = n_outputs * [None] if output_scale is None else output_scale
    output_offset = n_outputs * [None] if output_offset is None else output_offset
    output_data_range = n_outputs * [None] if output_data_range is None else output_data_range
    postprocessing = n_outputs * [None] if postprocessing is None else postprocessing
    halo = n_outputs * [None] if halo is None else halo

    outputs = [
        _get_output_tensor(root / test_out, name, reference, scale, offset, axes, data_range, postproc, hal)
        for test_out, name, reference, scale, offset, axes, data_range, postproc, hal in zip(
            test_outputs,
            output_names,
            output_reference,
            output_scale,
            output_offset,
            output_axes,
            output_data_range,
            postprocessing,
            halo,
        )
    ]

    # validate the pixel sizes (currently only used by deepimagej)
    spatial_axes = [[ax for ax in inp.axes if ax in "xyz"] for inp in inputs]
    if pixel_sizes is None:
        pixel_sizes = [{ax: 1.0 for ax in axes} for axes in spatial_axes]
    else:
        assert len(pixel_sizes) == n_inputs
        for pix_size, axes in zip(pixel_sizes, spatial_axes):
            assert isinstance(pix_size, dict)
            assert set(pix_size.keys()) == set(axes)

    #
    # generate general fields
    #
    format_version = get_args(model_spec.raw_nodes.FormatVersion)[-1]
    timestamp = datetime.datetime.now()

    authors = [model_spec.raw_nodes.Author(**a) for a in authors]
    cite = _build_cite(cite)
    documentation = _ensure_local(documentation, root)
    if covers is None:
        covers = _generate_covers(root / test_inputs[0], root / test_outputs[0], input_axes[0], output_axes[0], root)
    else:
        covers = _ensure_local(covers, root)
    if license is None:
        license = "CC-BY-4.0"

    # parse the weights
    weights, tmp_archtecture = _get_weights(
        weight_uri,
        weight_type,
        root,
        architecture,
        model_kwargs,
        tensorflow_version=tensorflow_version,
        opset_version=opset_version,
        pytorch_version=pytorch_version,
        dependencies=dependencies,
        attachments=weight_attachments,
    )

    # validate the sample inputs and outputs (if given)
    if sample_inputs is not None:
        assert sample_outputs is not None
        assert len(sample_inputs) == n_inputs
        assert len(sample_outputs) == n_outputs

    # add the deepimagej config if specified
    if add_deepimagej_config:
        if sample_inputs is None:
            sample_inputs, sample_outputs = _write_sample_data(
                test_inputs, test_outputs, input_axes, output_axes, pixel_sizes, root
            )
        # deepimagej expect tifs as sample data
        assert all(os.path.splitext(path)[1] in (".tif", ".tiff") for path in sample_inputs)
        assert all(os.path.splitext(path)[1] in (".tif", ".tiff") for path in sample_outputs)

        ij_config, ij_attachments = _get_deepimagej_config(
            root, test_inputs, test_outputs, input_axes, output_axes, pixel_sizes, preprocessing, postprocessing
        )

        if config is None:
            config = ij_config
        else:
            config.update(ij_config)

        if ij_attachments is not None:
            if attachments is None:
                attachments = {"files": ij_attachments}
            elif "files" not in attachments:
                attachments["files"] = ij_attachments
            else:
                attachments["files"] = list(set(attachments["files"]) | set(ij_attachments))

        if links is None:
            links = ["deepimagej/deepimagej"]
        else:
            links.append("deepimagej/deepimagej")

    # make sure links are unique
    if links is not None:
        links = list(set(links))

    # make sure sample inputs / outputs are relative paths
    if sample_inputs is not None:
        sample_inputs = _ensure_local_or_url(sample_inputs, root)

    if sample_outputs is not None:
        sample_outputs = _ensure_local_or_url(sample_outputs, root)

    # optional kwargs, don't pass them if none
    optional_kwargs = {
        "config": config,
        "git_repo": git_repo,
        "packaged_by": packaged_by,
        "run_mode": run_mode,
        "sample_inputs": sample_inputs,
        "sample_outputs": sample_outputs,
        "links": links,
    }
    kwargs = {k: v for k, v in optional_kwargs.items() if v is not None}

    if attachments is not None:
        kwargs["attachments"] = spec.rdf.raw_nodes.Attachments(**attachments)

    if maintainers is not None:
        kwargs["maintainers"] = [model_spec.raw_nodes.Maintainer(**m) for m in maintainers]

    if parent is not None:
        kwargs["parent"] = parent

    if training_data is not None:
        if "id" in training_data:
            msg = f"If training data is specified via 'id' no other keys are allowed, got {training_data}"
            assert len(training_data) == 1, msg
            kwargs["training_data"] = training_data
        else:
            if "type" not in training_data:
                training_data["type"] = "dataset"
            if "format_version" not in training_data:
                training_data["format_version"] = spec.dataset.format_version

    try:
        model = model_spec.raw_nodes.Model(
            authors=authors,
            cite=cite,
            covers=covers,
            description=description,
            documentation=documentation,
            format_version=format_version,
            inputs=inputs,
            license=license,
            name=name,
            outputs=outputs,
            root_path=root,
            tags=tags,
            test_inputs=test_inputs,
            test_outputs=test_outputs,
            timestamp=timestamp,
            weights=weights,
            **kwargs,
        )
        model_package = export_resource_package(model, output_path=output_path)
    except Exception as e:
        raise e
    finally:
        if tmp_archtecture is not None:
            os.remove(tmp_archtecture)

    model = load_raw_resource_description(model_package)
    return model



import dataclasses

import pytest

from bioimageio.core.prediction_pipeline._processing import KNOWN_PROCESSING
from bioimageio.core.prediction_pipeline._utils import FIXED

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore


@pytest.mark.parametrize(
    "proc",
    list(KNOWN_PROCESSING["pre"].values()) + list(KNOWN_PROCESSING["post"].values()),
)
def test_no_req_measures_for_mode_fixed(proc):
    # check if mode=fixed is valid for this proc
    for f in dataclasses.fields(proc):
        if f.name == "mode":
            break
    else:
        raise AttributeError("Processing is missing mode attribute")
    # mode is always annotated as literals (or literals of literals)
    valid_modes = get_args(f.type)
    for inner in get_args(f.type):
        valid_modes += get_args(inner)

    if FIXED not in valid_modes:
        return

    # we might be missing required kwargs. These have marshmallow.missing value as default
    # and raise a TypeError is in __post_init__()
    proc.__post_init__ = lambda self: None  # ignore missing kwargs

    proc_instance = proc(tensor_name="tensor_name", mode=FIXED)
    req_measures = proc_instance.get_required_measures()
    assert not req_measures



import abc
from dataclasses import dataclass
from typing import Iterable, List, Optional, Sequence, Tuple, Union

import xarray as xr
from marshmallow import missing

from bioimageio.core.resource_io import nodes
from bioimageio.spec.model import raw_nodes
from ._combined_processing import CombinedProcessing
from ._model_adapters import ModelAdapter, create_model_adapter
from ._stat_state import StatsState
from ._utils import ComputedMeasures, Sample, TensorName
from .. import load_resource_description
from ..resource_io.utils import resolve_raw_node


@dataclass
class NamedImplicitOutputShape:
    reference_input: TensorName = missing
    scale: List[Tuple[str, float]] = missing
    offset: List[Tuple[str, int]] = missing

    def __len__(self):
        return len(self.scale)


class PredictionPipeline(abc.ABC):
    """
    Represents model computation including preprocessing and postprocessing
    Note: Ideally use the PredictionPipeline as a context manager
    """

    @abc.abstractmethod
    def __enter__(self):
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    @abc.abstractmethod
    def forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        """
        Compute predictions
        """
        ...

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Name of the pipeline
        """
        ...

    @property
    @abc.abstractmethod
    def input_specs(self) -> List[nodes.InputTensor]:
        """
        specs of inputs
        """
        ...

    @property
    @abc.abstractmethod
    def output_specs(self) -> List[nodes.OutputTensor]:
        """
        specs of outputs
        """
        ...

    @abc.abstractmethod
    def load(self) -> None:
        """
        optional step: load model onto devices before calling forward if not using it as context manager
        """
        ...

    @abc.abstractmethod
    def unload(self) -> None:
        """
        free any device memory in use
        """
        ...


class _PredictionPipelineImpl(PredictionPipeline):
    def __init__(
        self,
        *,
        name: str,
        bioimageio_model: Union[nodes.Model, raw_nodes.Model],
        preprocessing: CombinedProcessing,
        postprocessing: CombinedProcessing,
        ipt_stats: StatsState,
        out_stats: StatsState,
        model: ModelAdapter,
    ) -> None:
        if bioimageio_model.run_mode:
            raise NotImplementedError(f"Not yet implemented inference for run mode '{bioimageio_model.run_mode.name}'")

        self._name = name
        if isinstance(bioimageio_model, nodes.Model):
            self._input_specs = bioimageio_model.inputs
            self._output_specs = bioimageio_model.outputs
        else:
            assert isinstance(bioimageio_model, raw_nodes.Model)
            self._input_specs = [resolve_raw_node(s, nodes) for s in bioimageio_model.inputs]
            self._output_specs = [resolve_raw_node(s, nodes) for s in bioimageio_model.outputs]

        self._preprocessing = preprocessing
        self._postprocessing = postprocessing
        self._ipt_stats = ipt_stats
        self._out_stats = out_stats
        self._model: ModelAdapter = model

    def __call__(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        return self.forward(*input_tensors)

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unload()
        return False

    @property
    def name(self):
        return self._name

    @property
    def input_specs(self):
        return self._input_specs

    @property
    def output_specs(self):
        return self._output_specs

    def predict(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        """Predict input_tensor with the model without applying pre/postprocessing."""
        return self._model.forward(*input_tensors)

    def apply_preprocessing(self, sample: Sample, computed_measures: ComputedMeasures) -> None:
        """apply preprocessing in-place, also updates given computed_measures"""
        self._ipt_stats.update_with_sample(sample)
        for mode, stats in self._ipt_stats.compute_measures().items():
            if mode not in computed_measures:
                computed_measures[mode] = {}
            computed_measures[mode].update(stats)

        self._preprocessing.apply(sample, computed_measures)

    def apply_postprocessing(self, sample: Sample, computed_measures: ComputedMeasures) -> None:
        """apply postprocessing in-place, also updates given computed_measures"""
        self._out_stats.update_with_sample(sample)
        for mode, stats in self._out_stats.compute_measures().items():
            if mode not in computed_measures:
                computed_measures[mode] = {}
            computed_measures[mode].update(stats)

        self._postprocessing.apply(sample, computed_measures)

    def forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        """Apply preprocessing, run prediction and apply postprocessing.
        Note: The preprocessing might change input_tensors in-pace.
        """
        input_sample = dict(zip([ipt.name for ipt in self.input_specs], input_tensors))
        computed_measures = {}
        self.apply_preprocessing(input_sample, computed_measures)

        prediction_tensors = self.predict(*list(input_sample.values()))
        prediction = dict(zip([out.name for out in self.output_specs], prediction_tensors))
        self.apply_postprocessing(prediction, computed_measures)

        return [prediction[tn] for tn in [out.name for out in self.output_specs]]

    def load(self):
        self._model.load()

    def unload(self):
        self._model.unload()


def create_prediction_pipeline(
    bioimageio_model: Union[nodes.Model, raw_nodes.Model],
    *,
    devices: Optional[Sequence[str]] = None,
    weight_format: Optional[str] = None,
    dataset_for_initial_statistics: Iterable[Sequence[xr.DataArray]] = tuple(),
    update_dataset_stats_after_n_samples: Optional[int] = None,
    update_dataset_stats_for_n_samples: int = float("inf"),
    model_adapter: Optional[ModelAdapter] = None,
) -> PredictionPipeline:
    """
    Creates prediction pipeline which includes:
    * computation of input statistics
    * preprocessing
    * model prediction
    * computation of output statistics
    * postprocessing
    """
    model_adapter: ModelAdapter = model_adapter or create_model_adapter(
        bioimageio_model=bioimageio_model, devices=devices, weight_format=weight_format
    )
    if isinstance(bioimageio_model, nodes.Model):
        ipts = bioimageio_model.inputs
        outs = bioimageio_model.outputs

    else:
        assert isinstance(bioimageio_model, raw_nodes.Model)
        ipts = [resolve_raw_node(s, nodes) for s in bioimageio_model.inputs]
        outs = [resolve_raw_node(s, nodes) for s in bioimageio_model.outputs]

    preprocessing = CombinedProcessing(ipts)

    def sample_dataset():
        for tensors in dataset_for_initial_statistics:
            yield dict(zip([ipt.name for ipt in bioimageio_model.inputs], tensors))

    ipt_stats = StatsState(
        preprocessing.required_measures,
        dataset=sample_dataset(),
        update_dataset_stats_after_n_samples=update_dataset_stats_after_n_samples,
        update_dataset_stats_for_n_samples=update_dataset_stats_for_n_samples,
    )
    postprocessing = CombinedProcessing(outs)
    out_stats = StatsState(
        postprocessing.required_measures,
        dataset=tuple(),
        update_dataset_stats_after_n_samples=0,
        update_dataset_stats_for_n_samples=ipt_stats.sample_count + update_dataset_stats_for_n_samples,
    )

    return _PredictionPipelineImpl(
        name=bioimageio_model.name,
        bioimageio_model=bioimageio_model,
        model=model_adapter,
        preprocessing=preprocessing,
        postprocessing=postprocessing,
        ipt_stats=ipt_stats,
        out_stats=out_stats,
    )



import pathlib
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Union

from marshmallow import missing
from marshmallow.utils import _Missing

from bioimageio.spec.model import raw_nodes as model_raw_nodes
from bioimageio.spec.rdf import raw_nodes as rdf_raw_nodes
from bioimageio.spec.collection import raw_nodes as collection_raw_nodes
from bioimageio.spec.shared import raw_nodes


@dataclass
class Node(raw_nodes.RawNode):
    pass


@dataclass
class ResourceDescription(Node, raw_nodes.ResourceDescription):
    pass


@dataclass
class URI(Node, raw_nodes.URI):
    pass


@dataclass
class ParametrizedInputShape(Node, raw_nodes.ParametrizedInputShape):
    pass


@dataclass
class ImplicitOutputShape(Node, raw_nodes.ImplicitOutputShape):
    pass


@dataclass
class Dependencies(Node, raw_nodes.Dependencies):
    file: pathlib.Path = missing


@dataclass
class CiteEntry(Node, rdf_raw_nodes.CiteEntry):
    pass


@dataclass
class Author(Node, model_raw_nodes.Author):
    pass


@dataclass
class Maintainer(Node, model_raw_nodes.Maintainer):
    pass


@dataclass
class Badge(Node, rdf_raw_nodes.Badge):
    pass


@dataclass
class RDF(rdf_raw_nodes.RDF, ResourceDescription):
    badges: Union[_Missing, List[Badge]] = missing
    covers: Union[_Missing, List[Path]] = missing


@dataclass
class CollectionEntry(Node, collection_raw_nodes.CollectionEntry):
    source: URI = missing


@dataclass
class LinkedDataset(Node, model_raw_nodes.LinkedDataset):
    pass


@dataclass
class ModelParent(Node, model_raw_nodes.ModelParent):
    pass


@dataclass
class Collection(collection_raw_nodes.Collection, RDF):
    pass


@dataclass
class RunMode(Node, model_raw_nodes.RunMode):
    pass


@dataclass
class Preprocessing(Node, model_raw_nodes.Preprocessing):
    pass


@dataclass
class Postprocessing(Node, model_raw_nodes.Postprocessing):
    pass


@dataclass
class InputTensor(Node, model_raw_nodes.InputTensor):
    axes: Tuple[str, ...] = missing

    def __post_init__(self):
        super().__post_init__()
        # raw node has string with single letter axes names. Here we use a tuple of string here (like xr.DataArray).
        self.axes = tuple(self.axes)


@dataclass
class OutputTensor(Node, model_raw_nodes.OutputTensor):
    axes: Tuple[str, ...] = missing

    def __post_init__(self):
        super().__post_init__()
        # raw node has string with single letter axes names. Here we use a tuple of string here (like xr.DataArray).
        self.axes = tuple(self.axes)


@dataclass
class ImportedSource(Node):
    factory: Callable

    def __call__(self, *args, **kwargs):
        return self.factory(*args, **kwargs)


@dataclass
class KerasHdf5WeightsEntry(Node, model_raw_nodes.KerasHdf5WeightsEntry):
    source: Path = missing


@dataclass
class OnnxWeightsEntry(Node, model_raw_nodes.OnnxWeightsEntry):
    source: Path = missing


@dataclass
class PytorchStateDictWeightsEntry(Node, model_raw_nodes.PytorchStateDictWeightsEntry):
    source: Path = missing
    architecture: Union[_Missing, ImportedSource] = missing


@dataclass
class TorchscriptWeightsEntry(Node, model_raw_nodes.TorchscriptWeightsEntry):
    source: Path = missing


@dataclass
class TensorflowJsWeightsEntry(Node, model_raw_nodes.TensorflowJsWeightsEntry):
    source: Path = missing


@dataclass
class TensorflowSavedModelBundleWeightsEntry(Node, model_raw_nodes.TensorflowSavedModelBundleWeightsEntry):
    source: Path = missing


@dataclass
class Attachments(Node, rdf_raw_nodes.Attachments):
    files: Union[_Missing, List[Path]] = missing
    unknown: Union[_Missing, Dict[str, Any]] = missing


WeightsEntry = Union[
    KerasHdf5WeightsEntry,
    OnnxWeightsEntry,
    PytorchStateDictWeightsEntry,
    TensorflowJsWeightsEntry,
    TensorflowSavedModelBundleWeightsEntry,
    TorchscriptWeightsEntry,
]


@dataclass
class Model(model_raw_nodes.Model, RDF):
    authors: List[Author] = missing
    maintainers: Union[_Missing, List[Maintainer]] = missing
    test_inputs: List[Path] = missing
    test_outputs: List[Path] = missing
    weights: Dict[model_raw_nodes.WeightsFormat, WeightsEntry] = missing



import os
import pytest


# todo: test with 'any_torch_model'
def test_onnx_converter(unet2d_multi_tensor_or_not, tmp_path):
    from bioimageio.core.weight_converter.torch.onnx import convert_weights_to_onnx

    out_path = tmp_path / "weights.onnx"
    ret_val = convert_weights_to_onnx(unet2d_multi_tensor_or_not, out_path, test_decimal=3)
    assert os.path.exists(out_path)
    if not pytest.skip_onnx:
        assert ret_val == 0  # check for correctness is done in converter and returns 0 if it passes



from dataclasses import dataclass, field, fields
from typing import Mapping, Optional, Sequence, Type, Union

import numpy as np
import xarray as xr

from bioimageio.core.statistical_measures import Mean, Measure, Percentile, Std
from bioimageio.spec.model.raw_nodes import PostprocessingName, PreprocessingName
from ._utils import ComputedMeasures, DatasetMode, FIXED, Mode, PER_DATASET, PER_SAMPLE, RequiredMeasures, SampleMode

try:
    from typing import Literal, get_args, TypedDict
except ImportError:
    from typing_extensions import Literal, get_args, TypedDict  # type: ignore


def _get_fixed(
    fixed: Union[float, Sequence[float]], tensor: xr.DataArray, axes: Optional[Sequence[str]]
) -> Union[float, xr.DataArray]:
    if axes is None:
        return fixed

    fixed_shape = tuple(s for d, s in tensor.sizes.items() if d not in axes)
    fixed_dims = tuple(d for d in tensor.dims if d not in axes)
    fixed = np.array(fixed).reshape(fixed_shape)
    return xr.DataArray(fixed, dims=fixed_dims)


TensorName = str

MISSING = "MISSING"


@dataclass
class Processing:
    """base class for all Pre- and Postprocessing transformations"""

    tensor_name: str
    # todo: in python>=3.10 we should use dataclasses.KW_ONLY instead of MISSING (see child classes) to make inheritance work properly
    computed_measures: ComputedMeasures = field(default_factory=dict)
    mode: Mode = FIXED

    def get_required_measures(self) -> RequiredMeasures:
        return {}

    def set_computed_measures(self, computed: ComputedMeasures):
        # check if computed contains all required measures
        for mode, req_per_mode in self.get_required_measures().items():
            for tn, req_per_tn in req_per_mode.items():
                comp_measures = computed.get(mode, {}).get(tn, {})
                for req_measure in req_per_tn:
                    if req_measure not in comp_measures:
                        raise ValueError(f"Missing required {req_measure} for {tn} {mode}.")

        self.computed_measures = computed

    def get_computed_measure(self, tensor_name: TensorName, measure: Measure, *, mode: Optional[Mode] = None):
        """helper to unpack self.computed_measures"""
        ret = self.computed_measures.get(mode or self.mode, {}).get(tensor_name, {}).get(measure)
        if ret is None:
            raise RuntimeError(f"Missing computed {measure} for {tensor_name} {mode}.")

        return ret

    def __call__(self, tensor: xr.DataArray) -> xr.DataArray:
        return self.apply(tensor)

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        """apply processing"""
        raise NotImplementedError

    def __post_init__(self):
        # validate common kwargs by their annotations
        for f in fields(self):
            # check MISSING
            if getattr(self, f.name) is MISSING:
                raise TypeError(f"missing required argument {f.name}")

            if f.name == "mode":
                # mode is always annotated as literals (or literals of literals)
                valid_modes = get_args(f.type)
                for inner in get_args(f.type):
                    valid_modes += get_args(inner)

                if self.mode not in valid_modes:
                    raise NotImplementedError(f"Unsupported mode {self.mode} for {self.__class__.__name__}")


#
# helpers
#
def ensure_dtype(tensor: xr.DataArray, *, dtype) -> xr.DataArray:
    """
    Convert array to a given datatype
    """
    return tensor.astype(dtype)


#
# Pre- and Postprocessing implementations
#


@dataclass
class Binarize(Processing):
    threshold: float = MISSING  # make dataclass inheritance work for py<3.10 by using an explicit MISSING value.

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor > self.threshold, dtype="float32")


@dataclass
class Clip(Processing):
    min: float = MISSING
    max: float = MISSING

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor.clip(min=self.min, max=self.max), dtype="float32")


@dataclass
class EnsureDtype(Processing):
    dtype: str = MISSING

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor, dtype=self.dtype)


@dataclass
class ScaleLinear(Processing):
    """scale the tensor with a fixed multiplicative and additive factor"""

    gain: Union[float, Sequence[float]] = MISSING
    offset: Union[float, Sequence[float]] = MISSING
    axes: Optional[Sequence[str]] = None

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        scale_axes = tuple(ax for ax in tensor.dims if (ax not in self.axes and ax != "b"))
        if scale_axes:
            gain = xr.DataArray(np.atleast_1d(self.gain), dims=scale_axes)
            offset = xr.DataArray(np.atleast_1d(self.offset), dims=scale_axes)
        else:
            gain = self.gain
            offset = self.offset

        return ensure_dtype(tensor * gain + offset, dtype="float32")

    def __post_init__(self):
        super().__post_init__()
        if self.axes is None:
            assert isinstance(self.gain, (int, float))
            assert isinstance(self.offset, (int, float))


@dataclass
class ScaleMeanVariance(Processing):
    ...


@dataclass
class ScaleRange(Processing):
    mode: Literal[SampleMode, DatasetMode] = PER_SAMPLE
    axes: Optional[Sequence[str]] = None
    min_percentile: float = 0.0
    max_percentile: float = 100.0
    eps: float = 1e-6
    reference_tensor: Optional[TensorName] = None

    def get_required_measures(self) -> RequiredMeasures:
        axes = None if self.axes is None else tuple(self.axes)
        measures = {Percentile(self.min_percentile, axes=axes), Percentile(self.max_percentile, axes=axes)}
        return {self.mode: {self.reference_tensor or self.tensor_name: measures}}

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        ref_name = self.reference_tensor or self.tensor_name
        axes = None if self.axes is None else tuple(self.axes)
        v_lower = self.get_computed_measure(ref_name, Percentile(self.min_percentile, axes=axes))
        v_upper = self.get_computed_measure(ref_name, Percentile(self.max_percentile, axes=axes))

        return ensure_dtype((tensor - v_lower) / (v_upper - v_lower + self.eps), dtype="float32")

    def __post_init__(self):
        super().__post_init__()
        self.axes = None if self.axes is None else tuple(self.axes)  # make sure axes is Tuple[str] or None


@dataclass
class Sigmoid(Processing):
    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return 1.0 / (1.0 + np.exp(-tensor))


@dataclass
class ZeroMeanUnitVariance(Processing):
    mode: Mode = PER_SAMPLE
    mean: Optional[Union[float, Sequence[float]]] = None
    std: Optional[Union[float, Sequence[float]]] = None
    axes: Optional[Sequence[str]] = None
    eps: float = 1.0e-6

    def get_required_measures(self) -> RequiredMeasures:
        if self.mode == FIXED:
            return {}
        else:
            axes = None if self.axes is None else tuple(self.axes)
            return {self.mode: {self.tensor_name: {Mean(axes=axes), Std(axes=axes)}}}

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        axes = None if self.axes is None else tuple(self.axes)
        if self.mode == FIXED:
            assert self.mean is not None and self.std is not None
            mean = _get_fixed(self.mean, tensor, axes)
            std = _get_fixed(self.std, tensor, axes)
        elif self.mode in (PER_SAMPLE, PER_DATASET):
            assert self.mean is None and self.std is None
            mean = self.get_computed_measure(self.tensor_name, Mean(axes), mode=self.mode)
            std = self.get_computed_measure(self.tensor_name, Std(axes), mode=self.mode)
        else:
            raise ValueError(self.mode)

        tensor = (tensor - mean) / (std + self.eps)
        return ensure_dtype(tensor, dtype="float32")


_KnownProcessing = TypedDict(
    "_KnownProcessing",
    dict(pre=Mapping[PreprocessingName, Type[Processing]], post=Mapping[PostprocessingName, Type[Processing]]),
)

KNOWN_PROCESSING: _KnownProcessing = dict(
    pre={
        "binarize": Binarize,
        "clip": Clip,
        "scale_linear": ScaleLinear,
        "scale_range": ScaleRange,
        "sigmoid": Sigmoid,
        "zero_mean_unit_variance": ZeroMeanUnitVariance,
    },
    post={
        "binarize": Binarize,
        "clip": Clip,
        "scale_linear": ScaleLinear,
        "scale_mean_variance": ScaleMeanVariance,
        "scale_range": ScaleRange,
        "sigmoid": Sigmoid,
        "zero_mean_unit_variance": ZeroMeanUnitVariance,
    },
)



import abc
from typing import List, Optional, Sequence, Type, Union

import xarray as xr

from bioimageio.core import load_resource_description
from bioimageio.core.resource_io import nodes

#: Known weight formats in order of priority
#: First match wins
from bioimageio.spec.model import raw_nodes

_WEIGHT_FORMATS = ["pytorch_state_dict", "tensorflow_saved_model_bundle", "torchscript", "onnx", "keras_hdf5"]


class ModelAdapter(abc.ABC):
    """
    Represents model *without* any preprocessing and postprocessing
    """

    def __init__(
        self, *, bioimageio_model: Union[nodes.Model, raw_nodes.Model], devices: Optional[Sequence[str]] = None
    ):
        self.bioimageio_model = self._prepare_model(bioimageio_model)
        self.default_devices = devices
        self.loaded = False

    @staticmethod
    def _prepare_model(bioimageio_model):
        """the (raw) model node is prepared (here: loaded as non-raw model node) for the model adapter to be ready
        for operation.
        Note: To write a model adapter that uses the raw model node one can overwrite this method.
        """
        if isinstance(bioimageio_model, nodes.Model):
            return bioimageio_model
        else:
            return load_resource_description(bioimageio_model)

    def __enter__(self):
        """load on entering context"""
        assert not self.loaded
        self.load()  # using default_devices
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """unload on exiting context"""
        assert self.loaded
        self.unload()
        return False

    def load(self, *, devices: Optional[Sequence[str]] = None) -> None:
        """
        Note: Use ModelAdapter as context to not worry about calling unload()!
        Load model onto devices. If devices is None, self.default_devices are chosen
        (which may be None as well, in which case a framework dependent default is chosen)
        """
        self._load(devices=devices or self.default_devices)
        self.loaded = True

    @abc.abstractmethod
    def _load(self, *, devices: Optional[Sequence[str]] = None) -> None:
        """
        Load model onto devices. If devices is None a framework dependent default is chosen
        """
        ...

    def forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        """
        Load model if unloaded/outside context; then run forward pass of model to get model predictions
        """
        if not self.loaded:
            self.load()

        assert self.loaded
        return self._forward(*input_tensors)

    @abc.abstractmethod
    def _forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        """
        Run forward pass of model to get model predictions
        Note: model is responsible converting it's data representation to
        xarray.DataArray
        """
        ...

    def unload(self):
        """
        Unload model from any devices, freeing their memory.
        Note: Use ModelAdapter as context to not worry about calling unload()!
        """
        # implementation of non-state-machine logic in _unload()
        assert self.loaded
        self._unload()
        self.loaded = False

    @abc.abstractmethod
    def _unload(self) -> None:
        """
        Unload model from any devices, freeing their memory.
        """
        ...


def get_weight_formats() -> List[str]:
    """
    Return list of supported weight types
    """
    return _WEIGHT_FORMATS.copy()


def create_model_adapter(
    *,
    bioimageio_model: Union[nodes.Model, raw_nodes.Model],
    devices=Optional[Sequence[str]],
    weight_format: Optional[str] = None,
) -> ModelAdapter:
    """
    Creates model adapter based on the passed spec
    Note: All specific adapters should happen inside this function to prevent different framework
    initializations interfering with each other
    """
    weights = bioimageio_model.weights
    weight_formats = get_weight_formats()

    if weight_format is not None:
        if weight_format not in weight_formats:
            raise ValueError(f"Weight format {weight_format} is not in supported formats {weight_formats}")
        weight_formats = [weight_format]

    for weight in weight_formats:
        if weight in weights:
            adapter_cls = _get_model_adapter(weight)
            return adapter_cls(bioimageio_model=bioimageio_model, devices=devices)

    raise RuntimeError(
        f"weight format {weight_format} not among formats listed in model: {list(bioimageio_model.weights.keys())}"
    )


def _get_model_adapter(weight_format: str) -> Type[ModelAdapter]:
    """
    Return adapter class based on the weight format
    Note: All specific adapters should happen inside this function to prevent different framework
    initializations interfering with each other
    """
    if weight_format == "pytorch_state_dict":
        from ._pytorch_model_adapter import PytorchModelAdapter

        return PytorchModelAdapter

    elif weight_format == "tensorflow_saved_model_bundle":
        from ._tensorflow_model_adapter import TensorflowModelAdapter

        return TensorflowModelAdapter

    elif weight_format == "onnx":
        from ._onnx_model_adapter import ONNXModelAdapter

        return ONNXModelAdapter

    elif weight_format == "torchscript":
        from ._torchscript_model_adapter import TorchscriptModelAdapter

        return TorchscriptModelAdapter

    elif weight_format == "keras_hdf5":
        # keras can either be installed as a separate package or used as part of tensorflow
        # we try to first import the keras model adapter using the separate package and,
        # if it is not available, try to load the one using tf
        try:
            from ._keras_model_adapter import KerasModelAdapter
        except ImportError:
            from ._tensorflow_model_adapter import KerasModelAdapter

        return KerasModelAdapter

    else:
        raise ValueError(f"Weight format {weight_format} is not supported.")



import pathlib

import pytest


def test_error_for_wrong_shape(stardist_wrong_shape):
    from bioimageio.core.resource_tests import test_model

    summary = test_model(stardist_wrong_shape)[-1]
    expected_error_message = (
        "Shape (1, 512, 512, 33) of test output 0 'output' does not match output shape description: "
        "ImplicitOutputShape(reference_tensor='input', "
        "scale=[1.0, 1.0, 1.0, 0.0], offset=[1.0, 1.0, 1.0, 33.0])."
    )
    assert summary["error"] == expected_error_message


def test_error_for_wrong_shape2(stardist_wrong_shape2):
    from bioimageio.core.resource_tests import test_model

    summary = test_model(stardist_wrong_shape2)[-1]
    expected_error_message = (
        "Shape (1, 512, 512, 1) of test input 0 'input' does not match input shape description: "
        "ParametrizedInputShape(min=[1, 80, 80, 1], step=[0, 17, 17, 0])."
    )
    assert summary["error"] == expected_error_message


def test_test_model(any_model):
    from bioimageio.core.resource_tests import test_model

    summary = test_model(any_model)
    assert all([s["status"] for s in summary])


def test_test_resource(any_model):
    from bioimageio.core.resource_tests import test_resource

    summary = test_resource(any_model)
    assert all([s["status"] for s in summary])


def test_validation_section_warning(unet2d_nuclei_broad_model, tmp_path: pathlib.Path):
    from bioimageio.core.resource_tests import test_resource
    from bioimageio.core import load_resource_description

    model = load_resource_description(unet2d_nuclei_broad_model)

    summary = test_resource(model)[2]
    assert summary["name"] == "Test documentation completeness."
    assert summary["warnings"] == {"documentation": "No '# Validation' (sub)section found."}
    assert summary["status"] == "passed"

    doc_with_validation = tmp_path / "doc.md"
    doc_with_validation.write_text("# Validation\nThis is a section about how to validate the model on new data")
    model.documentation = doc_with_validation
    summary = test_resource(model)[2]
    assert summary["name"] == "Test documentation completeness."
    assert summary["warnings"] == {}
    assert summary["status"] == "passed"


@pytest.mark.skipif(pytest.skip_torch, reason="requires torch")
def test_issue289():
    """test for failure case from https://github.com/bioimage-io/core-bioimage-io-python/issues/289"""
    import bioimageio.core
    from bioimageio.core.resource_tests import test_model

    doi = "10.5281/zenodo.6287342"
    model_resource = bioimageio.core.load_resource_description(doi)
    test_result = test_model(model_resource)
    assert all([t["status"] == "passed" for t in test_result])



import numpy as np
import xarray as xr

from bioimageio.core.resource_io import nodes


def test_postprocessing_dtype():
    from bioimageio.core.prediction_pipeline._combined_processing import CombinedProcessing

    shape = [3, 32, 32]
    axes = ("c", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)

    threshold = 0.5
    exp = xr.DataArray(np_data > threshold, dims=axes)

    for dtype in ("float32", "float64", "uint8", "uint16"):
        outputs = [
            nodes.OutputTensor(
                "out1",
                data_type=dtype,
                axes=axes,
                shape=shape,
                postprocessing=[nodes.Postprocessing("binarize", dict(threshold=threshold))],
            )
        ]
        com_proc = CombinedProcessing(outputs)

        sample = {"out1": data}
        com_proc.apply(sample, {})
        res = sample["out1"]
        assert np.dtype(res.dtype) == np.dtype(dtype)
        xr.testing.assert_allclose(res, exp.astype(dtype))



import os
import pathlib
from copy import deepcopy
from tempfile import TemporaryDirectory
from typing import Dict, Optional, Sequence, Union
from zipfile import ZIP_DEFLATED, ZipFile

from marshmallow import missing

from bioimageio import spec
from bioimageio.core.resource_io.nodes import ResourceDescription
from bioimageio.spec import load_raw_resource_description
from bioimageio.spec.shared import raw_nodes
from bioimageio.spec.shared.common import (
    BIOIMAGEIO_CACHE_PATH,
    BIOIMAGEIO_USE_CACHE,
    get_class_name_from_type,
    no_cache_tmp_list,
)
from bioimageio.spec.shared.raw_nodes import ResourceDescription as RawResourceDescription
from . import nodes
from .utils import resolve_raw_node, resolve_source

serialize_raw_resource_description = spec.io_.serialize_raw_resource_description
save_raw_resource_description = spec.io_.save_raw_resource_description


def load_resource_description(
    source: Union[RawResourceDescription, ResourceDescription, os.PathLike, str, dict, raw_nodes.URI],
    *,
    weights_priority_order: Optional[Sequence[str]] = None,  # model only
) -> ResourceDescription:
    """load a BioImage.IO resource description file (RDF).
    This includes some transformations for convenience, e.g. importing `source`.
    Use `load_raw_resource_description` to obtain a raw representation instead.

    Args:
        source: resource description file (RDF) or raw BioImage.IO resource
        weights_priority_order: If given only the first weights format present in the model resource is included
    Returns:
        BioImage.IO resource
    """
    source = deepcopy(source)
    if isinstance(source, ResourceDescription):
        return source

    raw_rd = load_raw_resource_description(source, update_to_format="latest")

    if raw_rd.type == "model" and weights_priority_order is not None:
        for wf in weights_priority_order:
            if wf in raw_rd.weights:
                raw_rd.weights = {wf: raw_rd.weights[wf]}
                break
        else:
            raise ValueError(f"Not found any of the specified weights formats {weights_priority_order}")

    rd: ResourceDescription = resolve_raw_node(raw_rd=raw_rd, nodes_module=nodes)
    assert isinstance(rd, getattr(nodes, get_class_name_from_type(raw_rd.type)))

    return rd


def get_local_resource_package_content(
    source: RawResourceDescription,
    weights_priority_order: Optional[Sequence[Union[str]]],
    update_to_format: Optional[str] = None,
) -> Dict[str, Union[pathlib.Path, str]]:
    """

    Args:
        source: raw resource description
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.
        update_to_format: update resource to specific major.minor format version; ignoring patch version.

    Returns:
        Package content of local file paths or text content keyed by file names.

    """
    raw_rd = load_raw_resource_description(source, update_to_format=update_to_format)
    package_content = spec.get_resource_package_content(raw_rd, weights_priority_order=weights_priority_order)

    local_package_content = {}
    for k, v in package_content.items():
        if isinstance(v, raw_nodes.URI):
            v = resolve_source(v, raw_rd.root_path)
        elif isinstance(v, pathlib.Path):
            v = raw_rd.root_path / v

        local_package_content[k] = v

    return local_package_content


def export_resource_package(
    source: Union[RawResourceDescription, os.PathLike, str, dict, raw_nodes.URI],
    *,
    compression: int = ZIP_DEFLATED,
    compression_level: int = 1,
    output_path: Optional[os.PathLike] = None,
    update_to_format: Optional[str] = None,
    weights_priority_order: Optional[Sequence[Union[str]]] = None,
) -> pathlib.Path:
    """Package a BioImage.IO resource as a zip file.

    Args:
        source: raw resource description, path, URI or raw data as dict
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        output_path: file path to write package to
        update_to_format: update resource to specific "major.minor" or "latest" format version; ignoring patch version.
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        path to zipped BioImage.IO package in BIOIMAGEIO_CACHE_PATH or 'output_path'
    """
    raw_rd = load_raw_resource_description(source, update_to_format=update_to_format)
    package_content = get_local_resource_package_content(
        raw_rd, weights_priority_order, update_to_format=update_to_format
    )
    if output_path is None:
        package_path = _get_tmp_package_path(raw_rd, weights_priority_order)
    else:
        package_path = output_path

    make_zip(package_path, package_content, compression=compression, compression_level=compression_level)
    return package_path


def _get_package_base_name(raw_rd: RawResourceDescription, weights_priority_order: Optional[Sequence[str]]) -> str:
    package_file_name = raw_rd.name
    if raw_rd.version is not missing:
        package_file_name += f"_{raw_rd.version}"

    package_file_name = package_file_name.replace(" ", "_").replace(".", "_")

    return package_file_name


def _get_tmp_package_path(raw_rd: RawResourceDescription, weights_priority_order: Optional[Sequence[str]]):
    if BIOIMAGEIO_USE_CACHE:
        package_file_name = _get_package_base_name(raw_rd, weights_priority_order)
        cache_folder = BIOIMAGEIO_CACHE_PATH / "packages"
        cache_folder.mkdir(exist_ok=True, parents=True)

        package_path = (cache_folder / package_file_name).with_suffix(".zip")
        max_cached_packages_with_same_name = 100
        for p in range(max_cached_packages_with_same_name):
            if package_path.exists():
                package_path = (cache_folder / f"{package_file_name}p{p}").with_suffix(".zip")
            else:
                break
        else:
            raise FileExistsError(
                f"Already caching {max_cached_packages_with_same_name} versions of {cache_folder / package_file_name}!"
            )
    else:
        tmp_dir = TemporaryDirectory()
        no_cache_tmp_list.append(tmp_dir)
        package_path = pathlib.Path(tmp_dir.name) / "file"

    return package_path


def make_zip(
    path: os.PathLike, content: Dict[str, Union[str, pathlib.Path]], *, compression: int, compression_level: int
) -> None:
    """Write a zip archive.

    Args:
        path: output path to write to.
        content: dict with archive names and local file paths or strings for text files.
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile

    """
    with ZipFile(path, "w", compression=compression, compresslevel=compression_level) as myzip:
        for arc_name, file_or_str_content in content.items():
            if isinstance(file_or_str_content, str):
                myzip.writestr(arc_name, file_or_str_content)
            else:
                myzip.write(file_or_str_content, arcname=arc_name)



from typing import List, Optional, Sequence, Union

from bioimageio.core.resource_io import nodes
from ._processing import EnsureDtype, KNOWN_PROCESSING, Processing
from ._utils import ComputedMeasures, PER_DATASET, PER_SAMPLE, RequiredMeasures, Sample

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


class CombinedProcessing:
    def __init__(self, tensor_specs: Union[List[nodes.InputTensor], List[nodes.OutputTensor]]):
        PRE: Literal["pre"] = "pre"
        POST: Literal["post"] = "post"
        proc_prefix: Optional[Literal["pre", "post"]] = None
        self._procs = []
        for t in tensor_specs:
            if isinstance(t, nodes.InputTensor):
                steps = t.preprocessing or []
                if proc_prefix is not None and proc_prefix != PRE:
                    raise ValueError(f"Invalid mixed input/output tensor specs: {tensor_specs}")

                proc_prefix = PRE
            elif isinstance(t, nodes.OutputTensor):
                steps = t.postprocessing or []
                if proc_prefix is not None and proc_prefix != POST:
                    raise ValueError(f"Invalid mixed input/output tensor specs: {tensor_specs}")

                proc_prefix = POST
            else:
                raise NotImplementedError(t)

            for step in steps:
                self._procs.append(KNOWN_PROCESSING[proc_prefix][step.name](tensor_name=t.name, **step.kwargs))

        # There is a difference between pre-and-postprocessing:
        # Pre-processing always returns float32, because its output is consumed by the model.
        # Post-processing, however, should return the dtype that is specified in the model spec.
        # todo: cast dtype for inputs before preprocessing? or check dtype?
        if proc_prefix == POST:
            for t in tensor_specs:
                self._procs.append(EnsureDtype(tensor_name=t.name, dtype=t.data_type))

        self.required_measures: RequiredMeasures = self._collect_required_measures(self._procs)
        if proc_prefix == POST and self.required_measures[PER_DATASET]:
            raise NotImplementedError("computing statistics for output tensors per dataset is not yet implemented")

        self.tensor_names = [t.name for t in tensor_specs]

    def apply(self, sample: Sample, computed_measures: ComputedMeasures) -> None:
        for proc in self._procs:
            proc.set_computed_measures(computed_measures)
            sample[proc.tensor_name] = proc.apply(sample[proc.tensor_name])

    @staticmethod
    def _collect_required_measures(proc: Sequence[Processing]) -> RequiredMeasures:
        ret: RequiredMeasures = {PER_SAMPLE: {}, PER_DATASET: {}}
        for p in proc:
            for mode, ms_per_mode in p.get_required_measures().items():
                for tn, ms_per_tn in ms_per_mode.items():
                    if tn not in ret[mode]:
                        ret[mode][tn] = set()

                    ret[mode][tn].update(ms_per_tn)

        return ret



import warnings
import zipfile
from typing import List, Optional

import numpy as np
import tensorflow as tf
import xarray as xr
from marshmallow import missing

from ._model_adapter import ModelAdapter

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


class TensorflowModelAdapterBase(ModelAdapter):
    weight_format: Literal["keras_hdf5", "tensorflow_saved_model_bundle"]

    def require_unzipped(self, weight_file):
        if zipfile.is_zipfile(weight_file):
            out_path = weight_file.with_suffix("")
            with zipfile.ZipFile(weight_file, "r") as f:
                f.extractall(out_path)
            return out_path
        return weight_file

    def _load_model(self, weight_file):
        weight_file = self.require_unzipped(weight_file)
        if self.use_keras_api:
            return tf.keras.models.load_model(weight_file)
        else:
            # NOTE in tf1 the model needs to be loaded inside of the session, so we cannot preload the model
            return str(weight_file)

    def _load(self, *, devices: Optional[List[str]] = None):
        model_tf_version = self.bioimageio_model.weights[self.weight_format].tensorflow_version
        if model_tf_version is missing:
            model_tf_version = None
        else:
            model_tf_version = (int(model_tf_version.major), int(model_tf_version.minor))

        tf_version = tf.__version__
        tf_major_and_minor = tuple(map(int, tf_version.split(".")))[:2]
        if model_tf_version is None:
            warnings.warn(
                "The model did not contain metadata about the tensorflow version used for training."
                f"Cannot check if it is compatible with tf {tf_version}. The prediction result may be wrong."
            )
        elif tuple(model_tf_version[:2]) != tf_major_and_minor:
            warnings.warn(
                f"Model tensorflow version {model_tf_version} does not match {tf_version}."
                "The prediction results may be wrong"
            )

        tf_major_ver = tf_major_and_minor[0]
        assert tf_major_ver in (1, 2)
        self.use_keras_api = tf_major_ver > 1 or self.weight_format == KerasModelAdapter.weight_format

        # TODO tf device management
        if devices is not None:
            warnings.warn(f"Device management is not implemented for tensorflow yet, ignoring the devices {devices}")

        weight_file = self.require_unzipped(self.bioimageio_model.weights[self.weight_format].source)
        self._model = self._load_model(weight_file)
        self._internal_output_axes = [tuple(out.axes) for out in self.bioimageio_model.outputs]

    # TODO currently we relaod the model every time. it would be better to keep the graph and session
    # alive in between of forward passes (but then the sessions need to be properly opened / closed)
    def _forward_tf(self, *input_tensors):
        input_keys = [ipt.name for ipt in self.bioimageio_model.inputs]
        output_keys = [out.name for out in self.bioimageio_model.outputs]

        # TODO read from spec
        tag = tf.saved_model.tag_constants.SERVING
        signature_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY

        graph = tf.Graph()
        with graph.as_default():
            with tf.Session(graph=graph) as sess:

                # load the model and the signature
                graph_def = tf.saved_model.loader.load(sess, [tag], self._model)
                signature = graph_def.signature_def

                # get the tensors into the graph
                in_names = [signature[signature_key].inputs[key].name for key in input_keys]
                out_names = [signature[signature_key].outputs[key].name for key in output_keys]
                in_tensors = [graph.get_tensor_by_name(name) for name in in_names]
                out_tensors = [graph.get_tensor_by_name(name) for name in out_names]

                # run prediction
                res = sess.run(dict(zip(out_names, out_tensors)), dict(zip(in_tensors, input_tensors)))
                # from dict to list of tensors
                res = [res[out] for out in out_names]

        return res

    def _forward_keras(self, input_tensors):
        tf_tensor = [tf.convert_to_tensor(ipt) for ipt in input_tensors]
        result = self._model.forward(*tf_tensor)
        if not isinstance(result, (tuple, list)):
            result = [result]

        return [r if isinstance(r, np.ndarray) else tf.make_ndarray(r) for r in result]

    def _forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        data = [ipt.data for ipt in input_tensors]
        if self.use_keras_api:
            result = self._forward_keras(*data)
        else:
            result = self._forward_tf(*data)

        return [xr.DataArray(r, dims=axes) for r, axes in zip(result, self._internal_output_axes)]

    def _unload(self) -> None:
        warnings.warn("Device management is not implemented for keras yet, cannot unload model")


class TensorflowModelAdapter(TensorflowModelAdapterBase):
    weight_format = "tensorflow_saved_model_bundle"


class KerasModelAdapter(TensorflowModelAdapterBase):
    weight_format = "keras_hdf5"



import os
import re
import traceback
import warnings
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy
import numpy as np
import xarray as xr
from marshmallow import ValidationError

from bioimageio.core import __version__ as bioimageio_core_version, load_resource_description
from bioimageio.core.common import TestSummary
from bioimageio.core.prediction import predict
from bioimageio.core.prediction_pipeline import create_prediction_pipeline
from bioimageio.core.resource_io.nodes import (
    ImplicitOutputShape,
    Model,
    ParametrizedInputShape,
    ResourceDescription,
    URI,
)
from bioimageio.core.resource_io.utils import SourceNodeChecker
from bioimageio.spec import __version__ as bioimageio_spec_version
from bioimageio.spec.model.raw_nodes import WeightsFormat
from bioimageio.spec.shared import resolve_source
from bioimageio.spec.shared.common import ValidationWarning
from bioimageio.spec.shared.raw_nodes import ResourceDescription as RawResourceDescription


def test_model(
    model_rdf: Union[URI, Path, str],
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[List[str]] = None,
    decimal: int = 4,
) -> List[TestSummary]:
    """Test whether the test output(s) of a model can be reproduced."""
    return test_resource(
        model_rdf, weight_format=weight_format, devices=devices, decimal=decimal, expected_type="model"
    )


def check_input_shape(shape: Tuple[int, ...], shape_spec) -> bool:
    if isinstance(shape_spec, list):
        if shape != tuple(shape_spec):
            return False
    elif isinstance(shape_spec, ParametrizedInputShape):
        assert len(shape_spec.min) == len(shape_spec.step)
        if len(shape) != len(shape_spec.min):
            return False
        min_shape = shape_spec.min
        step = shape_spec.step
        # check if the shape is valid for all dimension by seeing if it can be reached with an integer number of steps
        # NOTE we allow that the valid shape is reached using a different number of steps for each axis here
        # this is usually valid because dimensions are independent in neural networks
        is_valid = [(sh - minsh) % st == 0 if st > 0 else sh == minsh for sh, st, minsh in zip(shape, step, min_shape)]
        return all(is_valid)
    else:
        raise TypeError(f"Encountered unexpected shape description of type {type(shape_spec)}")

    return True


def check_output_shape(shape: Tuple[int, ...], shape_spec, input_shapes) -> bool:
    if isinstance(shape_spec, list):
        return shape == tuple(shape_spec)
    elif isinstance(shape_spec, ImplicitOutputShape):
        ref_tensor = shape_spec.reference_tensor
        if ref_tensor not in input_shapes:
            raise ValidationError(f"The reference tensor name {ref_tensor} is not in {input_shapes}")
        ipt_shape = numpy.array(input_shapes[ref_tensor])
        scale = numpy.array([0.0 if sc is None else sc for sc in shape_spec.scale])
        offset = numpy.array(shape_spec.offset)
        exp_shape = numpy.round_(ipt_shape * scale) + 2 * offset

        return shape == tuple(exp_shape)
    else:
        raise TypeError(f"Encountered unexpected shape description of type {type(shape_spec)}")


def _test_resource_urls(rd: ResourceDescription) -> TestSummary:
    assert isinstance(rd, ResourceDescription)
    with warnings.catch_warnings(record=True) as all_warnings:
        try:
            SourceNodeChecker(root_path=rd.root_path).visit(rd)
        except FileNotFoundError as e:
            error = str(e)
            tb = traceback.format_tb(e.__traceback__)
        else:
            error = None
            tb = None

    return dict(
        name="All URLs and paths available",
        status="passed" if error is None else "failed",
        error=error,
        traceback=tb,
        bioimageio_spec_version=bioimageio_spec_version,
        bioimageio_core_version=bioimageio_core_version,
        nested_errors=None,
        source_name=rd.id or rd.id or rd.name if hasattr(rd, "id") else rd.name,
        warnings={"SourceNodeChecker": [str(w.message) for w in all_warnings]} if all_warnings else {},
    )


def _test_model_documentation(rd: ResourceDescription) -> TestSummary:
    assert isinstance(rd, Model)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        doc_path: Path = resolve_source(rd.documentation, root_path=rd.root_path)
        doc = doc_path.read_text()
        wrn = ""
        if not re.match("#.*[vV]alidation", doc):
            wrn = "No '# Validation' (sub)section found."

        return dict(
            name="Test documentation completeness.",
            status="passed",
            error=None,
            traceback=None,
            bioimageio_spec_version=bioimageio_spec_version,
            bioimageio_core_version=bioimageio_core_version,
            source_name=rd.id or rd.name if hasattr(rd, "id") else rd.name,
            warnings={"documentation": wrn} if wrn else {},
        )


def _test_model_inference(model: Model, weight_format: str, devices: Optional[List[str]], decimal: int) -> TestSummary:
    error: Optional[str] = None
    tb: Optional = None
    with warnings.catch_warnings(record=True) as all_warnings:
        try:
            inputs = [np.load(str(in_path)) for in_path in model.test_inputs]
            expected = [np.load(str(out_path)) for out_path in model.test_outputs]

            assert len(inputs) == len(model.inputs)  # should be checked by validation
            input_shapes = {}
            for idx, (ipt, ipt_spec) in enumerate(zip(inputs, model.inputs)):
                if not check_input_shape(tuple(ipt.shape), ipt_spec.shape):
                    raise ValidationError(
                        f"Shape {tuple(ipt.shape)} of test input {idx} '{ipt_spec.name}' does not match "
                        f"input shape description: {ipt_spec.shape}."
                    )
                input_shapes[ipt_spec.name] = ipt.shape

            assert len(expected) == len(model.outputs)  # should be checked by validation
            for idx, (out, out_spec) in enumerate(zip(expected, model.outputs)):
                if not check_output_shape(tuple(out.shape), out_spec.shape, input_shapes):
                    error = (error or "") + (
                        f"Shape {tuple(out.shape)} of test output {idx} '{out_spec.name}' does not match "
                        f"output shape description: {out_spec.shape}."
                    )

            with create_prediction_pipeline(
                bioimageio_model=model, devices=devices, weight_format=weight_format
            ) as prediction_pipeline:
                results = predict(prediction_pipeline, inputs)

            if len(results) != len(expected):
                error = (error or "") + (
                    f"Number of outputs and number of expected outputs disagree: {len(results)} != {len(expected)}"
                )
            else:
                for res, exp in zip(results, expected):
                    try:
                        np.testing.assert_array_almost_equal(res, exp, decimal=decimal)
                    except AssertionError as e:
                        error = (error or "") + f"Output and expected output disagree:\n {e}"
        except Exception as e:
            error = str(e)
            tb = traceback.format_tb(e.__traceback__)

    return dict(
        name="reproduce test outputs from test inputs",
        status="passed" if error is None else "failed",
        error=error,
        traceback=tb,
        bioimageio_spec_version=bioimageio_spec_version,
        bioimageio_core_version=bioimageio_core_version,
        warnings=ValidationWarning.get_warning_summary(all_warnings),
        source_name=model.id or model.name,
    )


def _test_load_resource(
    rdf: Union[RawResourceDescription, ResourceDescription, URI, Path, str],
    weight_format: Optional[WeightsFormat] = None,
) -> Tuple[Optional[ResourceDescription], TestSummary]:
    if isinstance(rdf, (URI, os.PathLike)):
        source_name = str(rdf)
    elif isinstance(rdf, str):
        source_name = rdf[:120]
    else:
        source_name = rdf.id if hasattr(rdf, "id") else rdf.name

    main_test_warnings = []
    try:
        with warnings.catch_warnings(record=True) as all_warnings:
            rd = load_resource_description(
                rdf, weights_priority_order=None if weight_format is None else [weight_format]
            )

            main_test_warnings += list(all_warnings)
    except Exception as e:
        error: Optional[str] = str(e)
        tb: Optional = traceback.format_tb(e.__traceback__)
    else:
        error = None
        tb = None

    load_summary = dict(
        name="load resource description",
        status="passed" if error is None else "failed",
        error=error,
        traceback=tb,
        bioimageio_spec_version=bioimageio_spec_version,
        bioimageio_core_version=bioimageio_core_version,
        warnings={},
        source_name=source_name,
    )

    return rd, load_summary


def _test_expected_resource_type(rd: ResourceDescription, expected_type: str) -> TestSummary:
    has_expected_type = rd.type == expected_type
    return dict(
        name="has expected resource type",
        status="passed" if has_expected_type else "failed",
        error=None if has_expected_type else f"expected type {expected_type}, found {rd.type}",
        traceback=None,
        source_name=rd.id or rd.name if hasattr(rd, "id") else rd.name,
    )


def test_resource(
    rdf: Union[RawResourceDescription, ResourceDescription, URI, Path, str],
    *,
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[List[str]] = None,
    decimal: int = 4,
    expected_type: Optional[str] = None,
) -> List[TestSummary]:
    """Test RDF dynamically

    Returns: summary dict with keys: name, status, error, traceback, bioimageio_spec_version, bioimageio_core_version
    """
    rd, load_test = _test_load_resource(rdf, weight_format)
    tests: List[TestSummary] = [load_test]
    if rd is not None:
        if expected_type is not None:
            tests.append(_test_expected_resource_type(rd, expected_type))

        tests.append(_test_resource_urls(rd))

    if isinstance(rd, Model):
        tests.append(_test_model_documentation(rd))
        tests.append(_test_model_inference(rd, weight_format, devices, decimal))

    return tests


def debug_model(
    model_rdf: Union[RawResourceDescription, ResourceDescription, URI, Path, str],
    *,
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[List[str]] = None,
):
    """Run the model test and return dict with inputs, results, expected results and intermediates.

    Returns dict with tensors "inputs", "inputs_processed", "outputs_raw", "outputs", "expected" and "diff".
    """
    inputs: Optional = None
    inputs_processed: Optional = None
    outputs_raw: Optional = None
    outputs: Optional = None
    expected: Optional = None
    diff: Optional = None

    model = load_resource_description(
        model_rdf, weights_priority_order=None if weight_format is None else [weight_format]
    )
    if not isinstance(model, Model):
        raise ValueError(f"Not a bioimageio.model: {model_rdf}")

    prediction_pipeline = create_prediction_pipeline(
        bioimageio_model=model, devices=devices, weight_format=weight_format
    )
    inputs = [
        xr.DataArray(np.load(str(in_path)), dims=input_spec.axes)
        for in_path, input_spec in zip(model.test_inputs, model.inputs)
    ]

    inputs_processed, stats = prediction_pipeline.preprocess(*inputs)
    outputs_raw = prediction_pipeline.predict(*inputs_processed)
    outputs, _ = prediction_pipeline.postprocess(*outputs_raw, input_sample_statistics=stats)
    if isinstance(outputs, (np.ndarray, xr.DataArray)):
        outputs = [outputs]

    expected = [
        xr.DataArray(np.load(str(out_path)), dims=output_spec.axes)
        for out_path, output_spec in zip(model.test_outputs, model.outputs)
    ]
    if len(outputs) != len(expected):
        error = f"Number of outputs and number of expected outputs disagree: {len(outputs)} != {len(expected)}"
        print(error)
    else:
        diff = []
        for res, exp in zip(outputs, expected):
            diff.append(res - exp)

    return {
        "inputs": inputs,
        "inputs_processed": inputs_processed,
        "outputs_raw": outputs_raw,
        "outputs": outputs,
        "expected": expected,
        "diff": diff,
    }



import os
import shutil
from pathlib import Path
from typing import Union
from zipfile import ZipFile

import bioimageio.spec as spec
from bioimageio.core import load_resource_description

import tensorflow
from tensorflow import saved_model


# adapted from
# https://github.com/deepimagej/pydeepimagej/blob/master/pydeepimagej/yaml/create_config.py#L236
def _convert_tf1(keras_weight_path, output_path, input_name, output_name, zip_weights):
    def build_tf_model():
        keras_model = keras.models.load_model(keras_weight_path)

        builder = saved_model.builder.SavedModelBuilder(output_path)
        signature = saved_model.signature_def_utils.predict_signature_def(
            inputs={input_name: keras_model.input}, outputs={output_name: keras_model.output}
        )

        signature_def_map = {saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: signature}

        builder.add_meta_graph_and_variables(
            keras.backend.get_session(), [saved_model.tag_constants.SERVING], signature_def_map=signature_def_map
        )
        builder.save()

    try:
        # try to build the tf model with the keras import from tensorflow
        from tensorflow import keras
        build_tf_model()
    except Exception:
        # if the above fails try to export with the standalone keras
        import keras

        build_tf_model()

    if zip_weights:
        zipped_model = f"{output_path}.zip"
        # zip the weights
        file_paths = []
        for folder_names, subfolder, filenames in os.walk(os.path.join(output_path)):
            for filename in filenames:
                # create complete filepath of file in directory
                file_paths.append(os.path.join(folder_names, filename))

        with ZipFile(zipped_model, "w") as zip_obj:
            for f in file_paths:
                # Add file to zip
                zip_obj.write(f, os.path.relpath(f, output_path))

        try:
            shutil.rmtree(output_path)
        except Exception:
            print("TensorFlow bundled model was not removed after compression")
        print("TensorFlow model exported to", zipped_model)
    else:
        print("TensorFlow model exported to", output_path)
    return 0


def convert_weights_to_tensorflow_saved_model_bundle(
    model_spec: Union[str, Path, spec.model.raw_nodes.Model], output_path: Union[str, Path]
):
    """Convert model weights from format 'keras_hdf5' to 'tensorflow_saved_model_bundle'.

    Adapted from
    https://github.com/deepimagej/pydeepimagej/blob/5aaf0e71f9b04df591d5ca596f0af633a7e024f5/pydeepimagej/yaml/create_config.py

    Args:
        model_spec: location of the resource for the input bioimageio model
        output_path: where to save the tensorflow weights. This path must not exist yet.
    """
    tf_major_ver = int(tensorflow.__version__.split(".")[0])

    path_ = Path(output_path)
    if path_.suffix == ".zip":
        path_ = Path(os.path.splitext(path_)[0])
        zip_weights = True
    else:
        zip_weights = False

    if path_.exists():
        raise ValueError(f"The ouptut directory at {path_} must not exist.")

    model = load_resource_description(model_spec)
    assert "keras_hdf5" in model.weights
    weight_spec = model.weights["keras_hdf5"]
    weight_path = str(weight_spec.source)

    if weight_spec.tensorflow_version:
        model_tf_major_ver = int(weight_spec.tensorflow_version.major)
        if model_tf_major_ver != tf_major_ver:
            raise RuntimeError(f"Tensorflow major versions of model {model_tf_major_ver} is not {tf_major_ver}")

    if tf_major_ver == 1:
        if len(model.inputs) != 1 or len(model.outputs) != 1:
            raise NotImplementedError(
                "Weight conversion for models with multiple inputs or outputs is not yet implemented."
            )
        return _convert_tf1(weight_path, str(path_), model.inputs[0].name, model.outputs[0].name, zip_weights)
    else:
        raise NotImplementedError("Weight conversion for tensorflow 2 is not yet implemented.")
